# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for _common functions. This is a mix of pytest and unittests.
TODO: Future, move all tests to pytest.
"""

from __future__ import absolute_import, print_function

from datetime import datetime
import unittest
import pytest
import click
from mock import patch

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict  # pylint: disable=import-error

from pywbem import CIMClass, CIMProperty, CIMQualifier, CIMInstance, \
    CIMQualifierDeclaration, CIMInstanceName, Uint8, Uint32, Uint64, Sint32, \
    CIMDateTime, CIMClassName
from pywbem._nocasedict import NocaseDict
from pywbemtools.pywbemcli._common import parse_wbemuri_str, \
    filter_namelist, parse_kv_pair, split_array_value, sort_cimobjects, \
    create_ciminstance, compare_instances, resolve_propertylist, \
    _format_instances_as_rows, _print_instances_as_table, is_classname, \
    pick_one_from_list, pick_multiple_from_list, hide_empty_columns, \
    verify_operation, split_str_w_esc, format_keys
# pylint: disable=unused-import
from pywbemtools.pywbemcli._context_obj import ContextObj


from tests.unit.pytest_extensions import simplified_test_function
# from tests.unit.utils import assert_lines

DATETIME1_DT = datetime(2014, 9, 22, 10, 49, 20, 524789)
DATETIME1_OBJ = CIMDateTime(DATETIME1_DT)
DATETIME1_STR = '"20140922104920.524789+000"'

OK = True     # mark tests OK when they execute correctly
RUN = True    # Mark OK = False and current test case being created RUN
FAIL = False  # Any test currently FAILING or not tested yet
SKIP = False  # mark tests that are to be skipped.


TESTCASES_ISCLASSNAME = [
    # TESTCASES for resolve_propertylist
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * name: string containing classname or instancename
    #   * exp_rtn: expected function return.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    ('verify is a classname',
     dict(name=u"CIM_Blah", exp_rtn=True),
     None, None, True),

    ('verify is instance name',
     dict(name=u"CIM_Blah.px=3", exp_rtn=True),
     None, None, True),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_ISCLASSNAME)
@simplified_test_function
def test_is_classname(testcase, name, exp_rtn):
    """Test for resolve_propertylist function"""
    # The code to be tested

    act_rtn = is_classname(name)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_rtn == exp_rtn


TESTCASES_FORMAT_KEYBINDINGS = [
    # TESTCASES for resolve_propertylist
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * kb: keybinding
    #   * width - integer representing width of resulting field
    #   * exp_rtn: expected function return.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    ('Verify simple keybinding',
     dict(kb=NocaseDict([('kEY1', u'Ham')]),
          width=100,
          exp_rtn='kEY1="Ham"'),
     None, None, True),

    ('Verify multiple keys keybinding',
     dict(kb=NocaseDict([('kEY1', u'Ham'), ('key2', 3)]),
          width=100,
          exp_rtn='kEY1="Ham",key2=3'),
     None, None, True),

    ('Verify multiple keys binding with spaces in keys',
     dict(kb=NocaseDict([('kEY1', u'Ham and eggs'), ('key2', 'More eggs')]),
          width=100,
          exp_rtn='kEY1="Ham and eggs",key2="More eggs"'),
     None, None, True),

    ('Verify multiple keys binding multiple key types',
     dict(kb=NocaseDict([('Name', 'Foo'),
                         ('Number', Uint8(42)),
                         ('Boolean', False),
                         ('Ref', CIMInstanceName('CIM_Bar'))]),
          width=100,
          exp_rtn='Name="Foo",Number=42,Boolean=FALSE,Ref="/:CIM_Bar"'),
     None, None, True),

    ('Verify mutliple keys that fold into multiple lines',
     dict(kb=NocaseDict([('kEY1', u'Ham'), ('key2', 3)]),
          width=14,
          exp_rtn='kEY1="Ham"\nkey2=3'),
     None, None, True),

    ('Verify multiple keys binding with spaces in keys that fold',
     dict(kb=NocaseDict([('kEY1', u'Ham and eggs'), ('key2', 'More eggs')]),
          width=25,
          exp_rtn='kEY1="Ham and eggs"\nkey2="More eggs"'),
     None, None, True),

    ('Verify multiple keys binding with many keys keys without fold',
     dict(kb=NocaseDict([('k1', 1), ('k2', 2), ('k3', 3), ('k4', 4), ('k5', 5),
                         ('k6', 6), ('k7', 7), ('k8', 8)]),
          width=100,
          exp_rtn=('k1=1,k2=2,k3=3,k4=4,k5=5,k6=6,k7=7,k8=8')),
     None, None, True),

    ('Verify multiple keys binding with many keys every key folds',
     dict(kb=NocaseDict([('k1', 1), ('k2', 2), ('k3', 3), ('k4', 4), ('k5', 5),
                         ('k6', 6), ('k7', 7), ('k8', 8)]),
          width=3,
          exp_rtn=('k1=1\nk2=2\nk3=3\nk4=4\nk5=5\nk6=6\nk7=7\nk8=8')),
     None, None, True),


    ('Verify multiple keys binding with many keys keys some fold',
     dict(kb=NocaseDict([('k1', 1), ('k2', 2), ('k3', 3), ('k4', 4), ('k5', 5),
                         ('k6', 6), ('k7', 7), ('k8', 8)]),
          width=8,
          exp_rtn=('k1=1,k2=2\nk3=3,k4=4\nk5=5,k6=6\nk7=7,k8=8')),
     None, None, False),

    ('Verify multiple keys binding with spaces in keys',
     dict(kb=NocaseDict([('Name', 'Foo'),
                         ('Number', Uint8(42)),
                         ('Boolean', False),
                         ('Ref', CIMInstanceName('CIM_Bar'))]),
          width=4,
          exp_rtn='Name="Foo"\nNumber=42\nBoolean=FALSE\nRef="/:CIM_Bar"'),
     None, None, True),

    # Test no keys
    ('Verify no keys',
     dict(kb=NocaseDict(),
          width=100,
          exp_rtn=''),
     None, None, True),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_FORMAT_KEYBINDINGS)
@simplified_test_function
def test_format_keybindings(testcase, kb, width, exp_rtn):
    """Test for resolve_propertylist function"""
    # The code to be tested

    kbs = CIMInstanceName('blah', kb)
    act_rtn = format_keys(kbs, width)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_rtn == exp_rtn


TESTCASES_HIDE_EMPTY_COLUMNS = [
    # TESTCASES for hide_empty_column
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * rows: Array of strings (columns, rows)
    #   * headrs: list of strings
    #   * exp_rtn: expected function return.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    ('Verify deletes nothing with integers in rows',
     dict(rows=[[1, 2, 3],
                [4, 5, 6],
                [7, 8, 9]],
          headers=['h1', 'h2', 'h3'],
          exp_rtn={'rows': [[1, 2, 3],
                            [4, 5, 6],
                            [7, 8, 9]],
                   'headers': ['h1', 'h2', 'h3']}),
     None, None, True),

    ('Verify deletes nothing with integers in rows',
     dict(rows=[[1, 2, 0],
                [4, 5, 0],
                [7, 8, 0]],
          headers=['h1', 'h2', 'h3'],
          exp_rtn={'rows': [[1, 2, 0],
                            [4, 5, 0],
                            [7, 8, 0]],
                   'headers': ['h1', 'h2', 'h3']}),
     None, None, True),

    ('Verify deletes nothing with strings in rows',
     dict(rows=[['1', '2', '3'],
                ['4', '5', '6'],
                ['7', '8', '9']],
          headers=['h1', 'h2', 'h3'],
          exp_rtn={'rows': [['1', '2', '3'],
                            ['4', '5', '6'],
                            ['7', '8', '9']],
                   'headers': ['h1', 'h2', 'h3']}),
     None, None, True),

    ('Verify deletes nothing with strings in rows',
     dict(rows=[['1', '2', ''],
                ['4', '5', ''],
                ['7', '8', '']],
          headers=['h1', 'h2', 'h3'],
          exp_rtn={'rows': [['1', '2'],
                            ['4', '5'],
                            ['7', '8']],
                   'headers': ['h1', 'h2']}),
     None, None, True),

    ('Verify deletes nothing with strings in rows',
     dict(rows=[['1', '2', '3'],
                ['4', '5', ''],
                ['7', '8', '']],
          headers=['h1', 'h2', 'h3'],
          exp_rtn={'rows': [['1', '2', '3'],
                            ['4', '5', ''],
                            ['7', '8', '']],
                   'headers': ['h1', 'h2', 'h3']}),
     None, None, True),

    ('Verify deletes 3rd with strings in rows',
     dict(rows=[['1', '2', None],
                ['4', '5', None],
                ['7', '8', None]],
          headers=['h1', 'h2', 'h3'],
          exp_rtn={'rows': [['1', '2'],
                            ['4', '5'],
                            ['7', '8']],
                   'headers': ['h1', 'h2']}),
     None, None, True),

    ('Verify deletes 2nd with strings in rows',
     dict(rows=[['1', None, '3'],
                ['4', None, '6'],
                ['7', None, '9']],
          headers=['h1', 'h2', 'h3'],
          exp_rtn={'rows': [['1', '3'],
                            ['4', '6'],
                            ['7', '9']],
                   'headers': ['h1', 'h3']}),
     None, None, True),

    ('Verify deletes 1st 2nd with strings in rows',
     dict(rows=[[None, None, '3'],
                [None, None, '6'],
                [None, None, '9']],
          headers=['h1', 'h2', 'h3'],
          exp_rtn={'rows': [['3'],
                            ['6'],
                            ['9']],
                   'headers': ['h3']}),
     None, None, True),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_HIDE_EMPTY_COLUMNS)
@simplified_test_function
def test_hide_empty_columns(testcase, rows, headers, exp_rtn):
    """Test for resolve_propertylist function"""
    # The code to be tested

    act_rtn_headrs, act_rtn_rows = hide_empty_columns(headers, rows)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_rtn_rows == exp_rtn['rows']
    assert act_rtn_headrs == exp_rtn['headers']


TESTCASES_SPLIT_STR = [
    # TESTCASES for split_value_str
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * input_str: string to split
    #   * delimiter: split delimiter
    #   * exp_rtn: expected list of strings returned
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    ('Verify simple split',
     dict(input_str="abc,def,ijk",
          delimiter=',',
          exp_rtn=['abc', 'def', 'ijk']),
     None, None, True),

    ('Verify simple split empty entry',
     dict(input_str="abc,,ijk",
          delimiter=',',
          exp_rtn=['abc', '', 'ijk']),
     None, None, True),

    ('Verify string with escape',
     dict(input_str="abc,de\\,f,ijk",
          delimiter=',',
          exp_rtn=['abc', 'de,f', 'ijk']),
     None, None, True),

    ('Verify string with double escape',
     dict(input_str="abc,de\\,f,ijk",
          delimiter=',',
          exp_rtn=['abc', 'de,f', 'ijk']),
     None, None, True),

    ('Verify string with escape that should be ignored',
     dict(input_str="abc,de\\xf,ijk",
          delimiter=',',
          exp_rtn=['abc', 'de\\xf', 'ijk']),
     None, None, True),

    ('Verify string with trailingescape that should be ignored',
     dict(input_str="abc,def,ijk\\",
          delimiter=',',
          exp_rtn=['abc', 'def', 'ijk\\']),
     None, None, True),

    ('Verify string with leading escape that should be ignored',
     dict(input_str="\\abc,def,ijk",
          delimiter=',',
          exp_rtn=['\\abc', 'def', 'ijk']),
     None, None, True),

    ('Verify string with leading escape',
     dict(input_str="\\,abc,def,ijk",
          delimiter=',',
          exp_rtn=[',abc', 'def', 'ijk']),
     None, None, True),

    ('Verify string with multiple escape',
     dict(input_str="\\,abc\\,,def,\\,ijk",
          delimiter=',',
          exp_rtn=[',abc,', 'def', ',ijk']),
     None, None, True),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SPLIT_STR)
@simplified_test_function
def test_split_str(testcase, input_str, delimiter, exp_rtn):
    """Test for resolve_propertylist function"""
    # The code to be tested

    act_result = [item for item in split_str_w_esc(input_str, delimiter)]

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_result == exp_rtn


TESTCASES_PICK_ONE_FROM_LIST = [
    # TESTCASES for resolve_propertylist
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * options: tuple of strings defining properties
    #   * choices: list of choices to return from mock
    #   * exp_rtn: expected function return.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    ('Verify returns correct choice, in this case, ZERO',
     dict(options=[u'ZERO', u'ONE', u'TWO'], choices=['0'], exp_rtn=u'ZERO'),
     None, None, OK),

    ('Verify returns correct choice, in this case ONE',
     dict(options=[u'ZERO', u'ONE', u'TWO'], choices=['1'], exp_rtn=u'ONE'),
     None, None, OK),

    ('Verify returns correct choice, in this case ONE after one error',
     dict(options=[u'ZERO', u'ONE', u'TWO'], choices=['9', '1'],
          exp_rtn=u'ONE'),
     None, None, OK),

    ('Verify returns correct choice, in this case ONE after multipleerror',
     dict(options=[u'ZERO', u'ONE', u'TWO'], choices=['9', '-1', 'a', '2'],
          exp_rtn=u'TWO'),
     None, None, OK),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_PICK_ONE_FROM_LIST)
@simplified_test_function
def test_pick_one_from_list(testcase, options, choices, exp_rtn):
    """
    Test for pick_one_from_list function. Uses mock patch to define return
    values from the mock.
    """
    # setup mock for this test
    mock_function = 'pywbemtools.pywbemcli.click.prompt'
    with patch(mock_function, side_effect=choices) as mock_prompt:
        # The code to be tested
        title = "Test pick_one_from_list"
        context = ContextObj(None, None, None, None, None, None, None)
        act_rtn = pick_one_from_list(context, options, title)
        context.spinner.stop()
        assert mock_prompt.call_count == len(choices)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_rtn == exp_rtn


TESTCASES_PICK_MULTIPLE_FROM_LIST = [
    # TESTCASES for resolve_propertylist
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * options: tuple of strings defining properties
    #   * choices: item index from options that is to be chosen
    #   * exp_rtn: expected function return, a list of selected items.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # NOTE: choises must end with '' element to close the
    #       pick_multiple_from_list function.

    ('verify good choice ZERO made',
     dict(options=[u'ZERO', u'ONE', u'TWO'], choices=['0', ''],
          exp_rtn=[u'ZERO']),
     None, None, OK),

    ('verify good choice ONE made',
     dict(options=[u'ZERO', u'ONE', u'TWO'], choices=['1', ''],
          exp_rtn=[u'ONE']),
     None, None, OK),

    ('verify good choice TWO after bad choices',
     dict(options=[u'ZERO', u'ONE', u'TWO'],
          choices=['-1', '9', '3', 'a', '2', ''],
          exp_rtn=[u'TWO']),
     None, None, OK),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_PICK_MULTIPLE_FROM_LIST)
@simplified_test_function
def test_pick_multiple_from_list(testcase, options, choices, exp_rtn):
    """Test for pick_one_from_list function"""
    # setup mock for this test
    mock_function = 'pywbemtools.pywbemcli.click.prompt'
    with patch(mock_function, side_effect=choices) as mock_prompt:
        # The code to be tested
        title = "test_pick_multiple_from_list"
        # context = ContextObj(None, None, None, None, None, None, None)
        act_rtn = pick_multiple_from_list(None, options, title)
        # context.spinner.stop()
        assert mock_prompt.call_count == len(choices)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_rtn == exp_rtn


TESTCASES_RESOLVE_PROPERTYLIST = [
    # TESTCASES for resolve_propertylist
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * pl_str: tuple of strings defining properties
    #   * exp_pl: expected list return.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    ('verify simple property list with 2 entries',
     dict(pl_str=("abc,def",), exp_pl=['abc', 'def']),
     None, None, True),

    ('verify propertylist with single property entry',
     dict(pl_str=("abc",), exp_pl=['abc']),
     None, None, True),

    ('verify multiple properties',
     dict(pl_str=("abc", "def"), exp_pl=['abc', 'def']),
     None, None, True),

    ('verify multiple properties and both multiple in on option and multiple '
     'options.',
     dict(pl_str=None, exp_pl=None),
     None, None, True),

    ('verify multiple properties and both multiple in on option and multiple '
     'options.',
     dict(pl_str=("ab", "def", "xyz,rst"), exp_pl=['ab', 'def', 'xyz', 'rst']),
     None, None, True),


    ('verify empty propertylist',
     dict(pl_str=("",), exp_pl=[]),
     None, None, False),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_RESOLVE_PROPERTYLIST)
@simplified_test_function
def test_propertylist(testcase, pl_str, exp_pl):
    """Test for resolve_propertylist function"""
    # The code to be tested

    # wraps the test string in a tuple because that is the way the
    # propertylist option returns the list since it is a multiple type
    # option
    plist = resolve_propertylist(pl_str)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert plist == exp_pl


TESTCASES_COMPARE_INSTANCES = [
    # TESTCASES for resolve_propertylist
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * inst1: first instance for compare
    #   * inst2: second instance for compare
    #   * result: TODO
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    ('verify instances match',
     dict(inst1=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          inst2=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          result=True),
     None, None, True),

    ('verify classnames do not match',
     dict(inst1=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          inst2=CIMInstance('CIM_Foo1',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          result=False),
     None, None, True),

    ('verify property values do not match',
     dict(inst1=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          inst2=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo1', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          result=False),
     None, None, True),

    ('verify property names do not match',
     dict(inst1=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          inst2=CIMInstance('CIM_Foo',
                            properties={'Name1': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          result=False),
     None, None, True),

    ('verify classnames qualifiers do not match',
     dict(inst1=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          inst2=CIMInstance('CIM_Foo1',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          result=False),
     None, None, True),

    ('verify instances do not match diff  in number of properties',
     dict(inst1=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          inst2=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          result=False),
     None, None, True),

    ('verify instances values do not match',
     dict(inst1=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          inst2=CIMInstance('CIM_Foo',
                            properties={'Name': 'Foo', 'Chicken': 'Ham1'},
                            qualifiers={'Key': CIMQualifier('Key', True)},
                            path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'})),
          result=False),
     None, None, True),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_COMPARE_INSTANCES)
@simplified_test_function
def test_compare_instances(testcase, inst1, inst2, result):
    """Test for _common compare_instances function"""
    # The code to be tested

    rtn = compare_instances(inst1, inst2)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert rtn == result


# TODO: The mock for the following test is broken. Not sure yet how to
# fix it
TESTCASES_VERIFY_OPERATION = [
    # TESTCASES for resolve_propertylist
    #
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function and response:
    #   * txt: Text that would be fdisplayed
    #   * abort_msg: message that outputs with abort if response is n
    #   * result: True or False
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    ('Verify response y',
     dict(txt="blah",
          abort_msg=None,
          result=True),
     None, None, True),

    ('Verify response n',
     dict(txt="blahno",
          abort_msg=None,
          result=False),
     None, None, True),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_VERIFY_OPERATION)
@simplified_test_function
def test_verify_operation(testcase, txt, abort_msg, result):
    """
    This method mocks the click.confirm function to generate a response
    to the verify operation function
    """
    @patch('pywbemtools.pywbemcli.click.confirm', return_value=result)
    def test_verify_operation(txt, test_patch):
        return verify_operation(txt)

    # The code to be tested
    rtn = test_verify_operation(txt)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert rtn == result


# TODO: move this to pytest
class TestParseWbemUri(object):  # TODO: move this to pytest
    # pylint: disable=too-few-public-methods, useless-object-inheritance
    """
    Test CIMClassName.from_wbem_uri().
    """

    testcases = [
        # Testcases for CIMClassName.from_wbem_uri().
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * uri: WBEM URI string to be tested.
        # * ns: namespace parameter or None
        # * exp_result: Dict of all expected attributes of resulting object,
        #     if expected to succeed. Exception type, if expected to fail.
        # * exp_warn_type: Expected warning type.
        #     None, if no warning expected.
        # * condition: Condition for testcase to run.
        (
            "class and keys only case",
            '/root/cimv2:CIM_Foo.k1="v1"',
            None,
            dict(
                classname=u'CIM_Foo',
                namespace='root/cimv2',
                keys={'k1': 'v1'},
                host=None),
            None, True
        ),
        (
            "all components, normal case",
            'https://10.11.12.13:5989/root/cimv2:CIM_Foo.k1="v1"',
            None,
            dict(
                classname=u'CIM_Foo',
                namespace=u'root/cimv2',
                keys={'k1': 'v1'},
                host=u'10.11.12.13:5989'),
            None, True
        ),
        (
            "class and keybinding only",
            'CIM_Foo.k1="v1"',
            None,
            dict(
                classname=u'CIM_Foo',
                namespace=None,
                keys={'k1': 'v1'},
                host=None,),
            None, True
        ),
    ]

    @pytest.mark.parametrize(
        "desc, uri, ns, exp_result, exp_warn_type, condition",
        testcases)
    def test_parse_wbemuri_str(
            self, desc, uri, ns, exp_result, exp_warn_type, condition):
        # pylint: disable=unused-argument, no-self-use
        """Test function for parse_wbemuri_str."""

        if not condition:
            pytest.skip("Condition for test case not met")

        if isinstance(exp_result, type) and issubclass(exp_result, Exception):
            # We expect an exception
            exp_exc_type = exp_result
            exp_attrs = None
        else:
            # We expect the code to return
            exp_exc_type = None
            exp_attrs = exp_result

        if condition == 'pdb':
            import pdb
            pdb.set_trace()

        if exp_warn_type:
            with pytest.warns(exp_warn_type) as rec_warnings:
                if exp_exc_type:
                    with pytest.raises(exp_exc_type):
                        # The code to be tested
                        obj = parse_wbemuri_str(uri)

                else:
                    # The code to be tested
                    obj = parse_wbemuri_str(uri)

            assert len(rec_warnings) == 1

        else:
            if exp_exc_type:
                with pytest.raises(exp_exc_type):

                    # The code to be tested
                    obj = parse_wbemuri_str(uri)

            else:

                # The code to be tested
                obj = parse_wbemuri_str(uri)

        if exp_attrs:
            exp_classname = exp_attrs['classname']
            exp_namespace = exp_attrs['namespace']
            exp_host = exp_attrs['host']
            exp_keybindings = exp_attrs['keys']

            assert isinstance(obj, CIMInstanceName)

            assert obj.classname == exp_classname
            assert isinstance(obj.classname, type(exp_classname))

            assert obj.namespace == exp_namespace

            assert obj.keybindings == exp_keybindings

            assert obj.host == exp_host
            assert isinstance(obj.host, type(exp_host))


# TODO: Future pytestify this test and the others in this file
class FilterNamelistTest(object):  # pylint: disable=useless-object-inheritance
    """Test the common filter_namelist function."""

    name_list = ['CIM_abc', 'CIM_def', 'CIM_123', 'TST_abc']

    @pytest.mark.parametrize(
        "desc, regex, nl, match, ign_case",

        ['Verify TST_ case insensitive',
         'TST_', name_list, ['TST_abc'], True],

        ['Verify TSt_ case insensitive',
         'TSt_', name_list, ['TST_abc'], True],

        ['Verify TSt_ case insensitive',
         'TXST_', name_list, [], True],

        ['Verify TSt_ case insensitive',
         'CIM_', name_list, ['CIM_abc', 'CIM_def', 'CIM_123'], True],

        ['Verify TSt_ case sensitive',
         'TSt__', name_list, [], False],

        ['Verify wildcard filters',
         r'.*abc$', name_list, ['CIM_abc', 'TST_abc'], True],

        ['Verify wildcard filters',
         r'.*def', name_list, ['CIM_def'], True],
    )
    def test_filter_nameslist(self, desc, regex, nl, match, ign_case):
        # pylint: disable=no-self-use
        """
        Test filter_namelist function.
        """
        assert (filter_namelist(regex, nl, ignore_case=ign_case) == match), desc

    def test_case_insensitive(self):
        # pylint: disable=no-self-use
        """Test case insensitive match"""
        name_list = ['CIM_abc', 'CIM_def', 'CIM_123', 'TST_abc']

        assert filter_namelist('TST_', name_list) == ['TST_abc']
        assert filter_namelist('TSt_', name_list) == ['TST_abc']
        assert filter_namelist('XST_', name_list) == []
        assert filter_namelist('CIM_', name_list) == ['CIM_abc',
                                                      'CIM_def',
                                                      'CIM_123']

    def test_case_sensitive(self):
        # pylint: disable=no-self-use
        """Test case sensitive matches"""
        name_list = ['CIM_abc', 'CIM_def', 'CIM_123', 'TST_abc']

        assert filter_namelist('TSt_', name_list,
                               ignore_case=False) == []

    def test_wildcard_filters(self):
        # pylint: disable=no-self-use
        """Test more complex regex"""
        name_list = ['CIM_abc', 'CIM_def', 'CIM_123', 'TST_abc']
        assert filter_namelist(r'.*abc$', name_list) == ['CIM_abc',
                                                         'TST_abc']
        assert filter_namelist(r'.*def', name_list) == ['CIM_def']


# TODO: move this to pytest
class NameValuePairTest(object):
    # pylint: disable=too-few-public-methods, , useless-object-inheritance
    """Test simple name value pair parser"""

    def test_simple_pairs(self):
        # pylint: disable=no-self-use
        """Test simple pair parsing"""
        pname, value = parse_kv_pair('abc=test')
        assert pname == 'abc'
        assert value == 'test'

        pname, value = parse_kv_pair('abc=')
        assert pname == 'abc'
        assert value is None

        pname, value = parse_kv_pair('abc')
        assert pname == 'abc'
        assert value is None

        pname, value = parse_kv_pair('abc=12345')
        assert pname == 'abc'
        assert value == '12345'

        pname, value = parse_kv_pair('abc="fred"')
        assert pname == 'abc'
        assert value == '"fred"'

        pname, value = parse_kv_pair('abc="fr ed"')
        assert pname == 'abc'
        assert value == '"fr ed"'

        pname, value = parse_kv_pair('abc="fre\"d"')
        assert pname == 'abc'
        assert value == '"fre"d"'

        pname, value = parse_kv_pair('=def')
        assert pname == ''
        assert value == 'def'


class SorterTest(unittest.TestCase):
    """Test the object sort function in _common"""

    def test_sort_classes(self):
        """Test sorting list of classes"""

        classes = []

        classes.append(CIMClass(
            'CIM_Foo', properties={'InstanceID':
                                   CIMProperty('InstanceID', None,
                                               type='string')}))
        classes.append(CIMClass(
            'CIM_Boo', properties={'InstanceID':
                                   CIMProperty('InstanceID', None,
                                               type='string')}))
        classes.append(CIMClass(
            'CID_Boo', properties={'InstanceID':
                                   CIMProperty('InstanceID', None,
                                               type='string')}))
        sorted_rslt = sort_cimobjects(classes)
        self.assertEqual(len(classes), len(sorted_rslt))
        self.assertEqual(sorted_rslt[0].classname, 'CID_Boo')
        self.assertEqual(sorted_rslt[1].classname, 'CIM_Boo')
        self.assertEqual(sorted_rslt[2].classname, 'CIM_Foo')

        classes = []
        sorted_rslt = sort_cimobjects(classes)
        self.assertEqual(len(classes), len(sorted_rslt))

        classes = []
        sorted_rslt = sort_cimobjects(classes)
        classes.append(CIMClass(
            'CIM_Foo', properties={'InstanceID':
                                   CIMProperty('InstanceID', None,
                                               type='string')}))
        self.assertEqual(len(classes), len(sorted_rslt))
        self.assertEqual(sorted_rslt[0].classname, 'CIM_Foo')

    def test_sort_classnames(self):
        """Test sorting list of list of CIMClassName"""

        cln1 = CIMClassName("whoops", host="fred", namespace="root/cimv2")
        cln2 = CIMClassName("blah", host="fred", namespace="root/cimv2")
        cln3 = CIMClassName("blah", host="john", namespace="root/cimv2")
        cln4 = CIMClassName("blah", host="fred", namespace="root/cimv3")

        clns_in = [cln1, cln2, cln3, cln4]
        clns_exp = [cln2, cln1, cln4, cln3]

        sorted_rslt = sort_cimobjects(clns_in)

        self.assertEqual(len(clns_in), len(sorted_rslt))
        self.assertEqual(sorted_rslt, clns_exp)
        for obj in sorted_rslt:
            self.assertEqual(type(obj), CIMClassName)

    def test_sort_class_ref_rtn(self):
        """
        Test the return from the class references and class associators
        commands which is a tuple of CIMCLassName and CIMClass
        """

        cl1 = (CIMClass(
            'CIM_Foo', properties={'InstanceID':
                                   CIMProperty('InstanceID', None,
                                               type='string')}))
        cl2 = (CIMClass(
            'CIM_Boo', properties={'InstanceID':
                                   CIMProperty('InstanceID', None,
                                               type='string')}))
        cl3 = (CIMClass(
            'CID_Boo', properties={'InstanceID':
                                   CIMProperty('InstanceID', None,
                                               type='string')}))

        cln1 = CIMClassName("CIM_Foo", host="fred", namespace="root/cimv2")
        cln2 = CIMClassName("CIM_Boo", host="fred", namespace="root/cimv1")
        cln3 = CIMClassName("CIM_Boo", host="john", namespace="root/cimv2")

        input_tup = [(cln1, cl1), (cln2, cl2), (cln3, cl3)]
        sorted_rslt = sort_cimobjects(input_tup)
        self.assertEqual(len(input_tup), len(sorted_rslt))
        self.assertEqual(sorted_rslt[0][0], cln2)
        self.assertEqual(sorted_rslt[1][0], cln1)
        self.assertEqual(sorted_rslt[2][0], cln3)

    def test_sort_instancenames(self):
        """Test ability to sort list of instance names"""

        inst_names = []

        kb = {'Chicken': 'Ham', 'Beans': 42}
        obj = CIMInstanceName('CIM_Foo', kb)
        inst_names.append(obj)

        kb = {'Chicken': 'Ham', 'Beans': 42}
        obj = CIMInstanceName('CIM_Boo', kb)
        inst_names.append(obj)

        kb = {'Chicken': 'Ham', 'Beans': 42}
        obj = CIMInstanceName('CID_Foo', kb)
        inst_names.append(obj)

        sorted_rslt = sort_cimobjects(inst_names)
        self.assertEqual(len(inst_names), len(sorted_rslt))
        self.assertEqual(sorted_rslt[0].classname, 'CID_Foo')
        self.assertEqual(sorted_rslt[1].classname, 'CIM_Boo')
        self.assertEqual(sorted_rslt[2].classname, 'CIM_Foo')

    def test_sort_instances(self):
        """Test sort of instances"""
        instances = []

        props = {'Chicken': CIMProperty('Chicken', 'Ham'),
                 'Number': CIMProperty('Number', Uint32(42))}
        quals = {'Key': CIMQualifier('Key', True)}
        path = CIMInstanceName('CIM_Foo', {'Chicken': 'Ham'})

        obj = CIMInstance('CIM_Foo',
                          properties=props,
                          qualifiers=quals,
                          path=path)
        instances.append(obj)
        props = {'Chicken': CIMProperty('Chicken', 'Ham'),
                 'Number': CIMProperty('Number', Uint32(42))}
        quals = {'Key': CIMQualifier('Key', True)}
        path = CIMInstanceName('CIM_Boo', {'Chicken': 'Ham'})

        obj = CIMInstance('CIM_Boo',
                          properties=props,
                          qualifiers=quals,
                          path=path)
        instances.append(obj)
        instances_sorted = sort_cimobjects(instances)
        self.assertEqual(len(instances), len(instances_sorted))
        self.assertEqual(instances_sorted[0].classname, 'CIM_Boo')
        # TODO create multiple and test result

    def test_sort_qualifierdecls(self):
        """Test ability to sort list of qualifier declaractions"""

        qual_decls = []

        obj = CIMQualifierDeclaration('FooQualDecl3', 'uint32')
        qual_decls.append(obj)

        obj = CIMQualifierDeclaration('FooQualDecl2', 'uint32')
        qual_decls.append(obj)

        obj = CIMQualifierDeclaration('FooQualDecl1', 'uint32')
        qual_decls.append(obj)

        sorted_rslt = sort_cimobjects(qual_decls)
        self.assertEqual(len(qual_decls), len(sorted_rslt))
        self.assertEqual(sorted_rslt[0].name, 'FooQualDecl1')
        self.assertEqual(sorted_rslt[1].name, 'FooQualDecl2')
        self.assertEqual(sorted_rslt[2].name, 'FooQualDecl3')

    def test_sort_stringss(self):
        """Test ability to sort list of qualifier declaractions"""
        inputs = ['xyz', 'abc']
        sorted_rslt = sort_cimobjects(inputs)
        self.assertEqual(sorted_rslt, ['abc', 'xyz'])


# TODO: move this to pytest
class SplitTestNone(object):  # pylint: disable=useless-object-inheritance
    """Test splitting input parameters"""

    def split_test(self, input_str, exp_result):
        # pylint: disable=no-self-use
        """
        Common function to do the split and compare input to expected
        result.
        """
        act_result = split_array_value(input_str, ',')
        # print('split input %s result %s' % (input_str, result))
        assert (exp_result == act_result) % \
            'Failed split test exp %r, act %r' % (exp_result, act_result)

    def test_split(self):
        # pylint: disable=no-self-use
        """Define strings to split and call test function"""
        self.split_test('0,1,2,3,4,5,6',
                        ['0', '1', '2', '3', '4', '5', '6'])

        self.split_test('abc,def,jhi,klm,nop',
                        ['abc', 'def', 'jhi', 'klm', 'nop'])

        self.split_test('abc,def,jhi,klm,n\\,op',
                        ['abc', 'def', 'jhi', 'klm', 'n\\,op'])

        # self.split_test('abc,de f', ['abc','de f'])


# TODO: move this to pytest
class KVPairParsingTest(object):  # pylint: disable=useless-object-inheritance
    "Test parsing key/value pairs on input"

    def execute_test(self, test_string, exp_name, exp_value):
        # pylint: disable=no-self-use
        """Execute the function and test result"""

        act_name, act_value = parse_kv_pair(test_string)

        assert (exp_name == act_name), \
            ' KVPairParsing. Expected ' \
            ' name=%s, function returned %s' % (exp_name, act_name)
        assert (exp_value == act_value), \
            ' KVPairParsing. Expected ' \
            ' value=%s, act value=%s' % (exp_value, act_value)

    def test_scalar_int(self):
        # pylint: disable=no-self-use
        """Test for scalar integer value"""
        self.execute_test('prop_name=1', 'prop_name', str(1))
        self.execute_test('prop_name=91999', 'prop_name', str(91999))


# TODO; Convert this test to pytest
class CreateCIMInstanceTest(unittest.TestCase):
    """Test the function that creates a CIMInstance from cli args"""

    @staticmethod
    def create_scalar_class():
        """
        Create and return a class of scalar properties of all types
        except embedded instance and return the class
        """
        cls = CIMClass(
            'CIM_Foo', properties={'ID': CIMProperty('ID', None,
                                                     type='string'),
                                   'Boolp': CIMProperty('Boolp', None,
                                                        type='boolean'),
                                   'Uint8p': CIMProperty('Uint8p', None,
                                                         type='uint8'),
                                   'Sint8p': CIMProperty('Sint8p', None,
                                                         type='sint8'),
                                   'Uint16p': CIMProperty('Uint16p', None,
                                                          type='uint16'),
                                   'Sint16p': CIMProperty('Sint16p', None,
                                                          type='sint16'),
                                   'Uint32p': CIMProperty('Uint32p', None,
                                                          type='uint32'),
                                   'Sint32p': CIMProperty('Sint32p', None,
                                                          type='sint32'),
                                   'Uint64p': CIMProperty('Uint64p', None,
                                                          type='uint64'),
                                   'Sint64p': CIMProperty('Sint64p', None,
                                                          type='sint64'),
                                   'Real32p': CIMProperty('Real32p', None,
                                                          type='real32'),
                                   'Real64p': CIMProperty('Real64p', None,
                                                          type='real64'),
                                   'Dtp': CIMProperty('Dtp', None,
                                                      type='datetime'),
                                   'Strp': CIMProperty('Strp', None,
                                                       type='string')})
        return cls

    @staticmethod
    def create_array_class():
        """
        Create and return a class of scalar properties of all types
        except embedded instance and return the class
        """
        cls = CIMClass(
            'CIM_Foo', properties={'ID': CIMProperty('ID', None,
                                                     type='string'),
                                   'Boolp': CIMProperty('Boolp', None,
                                                        is_array=True,
                                                        type='boolean'),
                                   'Uint8p': CIMProperty('Uint8p', None,
                                                         is_array=True,
                                                         type='uint8'),
                                   'Sint8p': CIMProperty('Sint8p', None,
                                                         is_array=True,
                                                         type='sint8'),
                                   'Uint16p': CIMProperty('Uint16p', None,
                                                          is_array=True,
                                                          type='uint16'),
                                   'Sint16p': CIMProperty('Sint16p', None,
                                                          is_array=True,
                                                          type='sint16'),
                                   'Uint32p': CIMProperty('Uint32p', None,
                                                          is_array=True,
                                                          type='uint32'),
                                   'Sint32p': CIMProperty('Sint32p', None,
                                                          is_array=True,
                                                          type='sint32'),
                                   'Uint64p': CIMProperty('Uint64p', None,
                                                          is_array=True,
                                                          type='uint64'),
                                   'Sint64p': CIMProperty('Sint64p', None,
                                                          is_array=True,
                                                          type='sint64'),
                                   'Real32p': CIMProperty('Real32p', None,
                                                          is_array=True,
                                                          type='real32'),
                                   'Real64p': CIMProperty('Real64p', None,
                                                          is_array=True,
                                                          type='real64'),
                                   'Dtp': CIMProperty('Dtp', None,
                                                      is_array=True,
                                                      type='datetime'),
                                   'Strp': CIMProperty('Strp', None,
                                                       is_array=True,
                                                       type='string')})
        return cls

    def test_simple_scalar_instance(self):
        """Test scalar with one property"""
        cls = CreateCIMInstanceTest.create_scalar_class()

        exp_properties = {'ID': CIMProperty('ID', 'Testid', type='string')}
        exp_inst = CIMInstance('CIM_Foo',
                               properties=exp_properties)
        kv_properties = ['ID=Testid']
        act_inst = create_ciminstance(cls, kv_properties)
        # pp(act_inst)
        self.assertEqual(exp_inst, act_inst)

    # TODO add datetime property
    def test_scalar_instance(self):
        """
        Creates an instance from cmd line parameters and tests against
        predefined instance
        """
        cls = CreateCIMInstanceTest.create_scalar_class()

        exp_properties = {'ID': CIMProperty('ID', 'Testid', type='string'),
                          'Boolp': CIMProperty('Boolp', True),
                          'Uint8p': CIMProperty('Uint8p', 220, type='uint8'),
                          'Sint8p': CIMProperty('Sint8p', -120, type='sint8'),
                          'Uint32p': CIMProperty('Uint32p', 999, type='uint32'),
                          'Sint32p': CIMProperty('Sint32p', -99, type='sint32'),
                          'Uint64p': CIMProperty('Uint64p', 999, type='uint64'),
                          'Sint64p': CIMProperty('Sint64p', -99, type='sint64'),

                          'Strp': CIMProperty('Strp', 'hoho', type='string')}

        exp_inst = CIMInstance('CIM_Foo',
                               properties=exp_properties)

        kv_properties = ['ID=Testid', 'Boolp=true', 'Uint8p=220', 'Sint8p=-120',
                         'Uint32p=999', 'Sint32p=-99', 'Uint64p=999',
                         'Sint64p=-99', 'Strp=hoho']
        act_inst = create_ciminstance(cls, kv_properties)

        self.assertTrue(compare_instances(exp_inst, act_inst))

        self.assertEqual(exp_inst, act_inst)

    def test_simple_scalar_two_prop(self):
        """Test scalar with two property"""
        cls = CreateCIMInstanceTest.create_scalar_class()

        exp_properties = {'ID': CIMProperty('ID', 'Testid', type='string'),
                          'Boolp': CIMProperty('Boolp', False)}
        exp_inst = CIMInstance('CIM_Foo',
                               properties=exp_properties)
        kv_properties = ['ID=Testid', 'Boolp=false']
        act_inst = create_ciminstance(cls, kv_properties)
        # pp(act_inst)
        self.assertTrue(compare_instances(exp_inst, act_inst))
        self.assertEqual(exp_inst, act_inst)

    # TODO expand this to test errors for other types
    def test_simple_scalar_type_err(self):
        """Test scalar with type error"""
        cls = CreateCIMInstanceTest.create_scalar_class()

        try:
            kv_properties = ['Boolp=123']
            create_ciminstance(cls, kv_properties)
            self.fail('Expected exception to create_instance property type err')
        except click.ClickException:
            pass

        try:
            kv_properties = ['uint32p=shouldnotbestring']
            create_ciminstance(cls, kv_properties)
            self.fail('Expected exception to create_instance property with '
                      'value err')
        except click.ClickException:
            pass

    def test_simple_scalar_real(self):
        """Test scalar with two property"""
        cls = CreateCIMInstanceTest.create_scalar_class()

        exp_properties = {'ID': CIMProperty('ID', 'Testid', type='string'),
                          'Real32p': CIMProperty('Real32p', 1.99,
                                                 type='real32')}
        exp_inst = CIMInstance('CIM_Foo',
                               properties=exp_properties)
        kv_properties = ['ID=Testid', 'Real32p=1.99']
        act_inst = create_ciminstance(cls, kv_properties)
        self.assertTrue(compare_instances(exp_inst, act_inst))
        self.assertEqual(exp_inst, act_inst)

    # TODO test no properties, test property with no value

    def test_array_small_instance(self):
        """
        Creates an instance from cmd line parameters and tests against
        predefined instance
        """
        cls = CreateCIMInstanceTest.create_array_class()

        exp_properties = {'ID': CIMProperty('ID', 'Testid', type='string'),
                          'Boolp': CIMProperty('Boolp', [True, False])}

        exp_inst = CIMInstance('CIM_Foo',
                               properties=exp_properties)

        kv_properties = ['ID=Testid', 'Boolp=true,false']
        act_inst = create_ciminstance(cls, kv_properties)

        self.assertEqual(exp_inst, act_inst)

    # TODO add datetime property
    def test_array_instance(self):
        """
        Creates an instance from cmd line parameters and tests against
        predefined instance
        """
        cls = CreateCIMInstanceTest.create_array_class()

        exp_properties = {'ID': CIMProperty('ID', 'Testid', type='string'),
                          'Boolp': CIMProperty('Boolp', [True, False]),
                          'Uint8p': CIMProperty('Uint8p',
                                                [0, 12, 120], type='uint8'),
                          'Sint8p': CIMProperty('Sint8p',
                                                [-120, 0, 119], type='sint8'),
                          'Uint32p': CIMProperty('Uint32p',
                                                 [0, 999], type='uint32'),
                          'Sint32p': CIMProperty('Sint32p',
                                                 [-99, 0, 9999], type='sint32'),
                          'Uint64p': CIMProperty('Uint64p', [0, 999, 99999],
                                                 type='uint64'),
                          'Sint64p': CIMProperty('Sint64p', [-99, 0, 12345],
                                                 type='sint64'),
                          'Strp': CIMProperty('Strp', ['hoho', 'haha'],
                                              type='string')}
        exp_inst = CIMInstance('CIM_Foo',
                               properties=exp_properties)
        kv_properties = ['ID=Testid', 'Boolp=true,false', 'Uint8p=0,12,120',
                         'Sint8p=-120,0,119',
                         'Uint32p=0,999', 'Sint32p=-99,0,9999',
                         'Uint64p=0,999,99999',
                         'Sint64p=-99,0,12345', 'Strp=hoho,haha']

        act_inst = create_ciminstance(cls, kv_properties)
        # compare_instances(exp_inst, act_inst)
        self.assertEqual(exp_inst, act_inst,
                         'test_array_instance failed compare')

    def test_scalar_no_value(self):
        """Test scalar with one property with no value component"""
        cls = CreateCIMInstanceTest.create_array_class()

        exp_properties = {'ID': CIMProperty('ID', None, type='string')}
        exp_inst = CIMInstance('CIM_Foo',
                               properties=exp_properties)
        kv_properties = ['ID=']
        act_inst = create_ciminstance(cls, kv_properties)
        self.assertTrue(compare_instances(exp_inst, act_inst))
        self.assertEqual(exp_inst, act_inst)

    def test_scalar_empty_str(self):
        """Test scalar with one property where value is empty string"""
        cls = CreateCIMInstanceTest.create_array_class()

        exp_properties = {'ID': CIMProperty('ID', '""', type='string')}
        exp_inst = CIMInstance('CIM_Foo',
                               properties=exp_properties)
        kv_properties = ['ID=""']
        act_inst = create_ciminstance(cls, kv_properties)
        self.assertTrue(compare_instances(exp_inst, act_inst))
        self.assertEqual(exp_inst, act_inst)

    def test_invalid_propname(self):
        """Test scalar where input is property name not in class"""
        cls = CreateCIMInstanceTest.create_scalar_class()

        try:
            kv_properties = ['Boolpxxx=True']
            create_ciminstance(cls, kv_properties)
            self.fail('Expected exception to create_instance property type err')
        except click.ClickException:
            pass


# TODO this is a pytest. param
def simple_instance(pvalue=None):
    """
    Build a simple instance to test and return that instance. If the param
    pvalue is provided it is a scalar value and the instance with just
    this property is returned.
    """
    if pvalue:
        properties = [CIMProperty("P", pvalue)]
    else:
        properties = [CIMProperty("Pbt", value=True),
                      CIMProperty("Pbf", value=False),
                      CIMProperty("Pint32", Uint32(99)),
                      CIMProperty("Pint64", Uint64(9999)),
                      CIMProperty("Pdt", DATETIME1_OBJ),
                      CIMProperty("Pstr1", u"Test String")]
    inst = CIMInstance("CIM_Foo", properties)
    return inst


def simple_instance2(pvalue=None):
    """
    Build a simple instance to test and return that instance. If the param
    pvalue is provided it is a scalar value and the instance with just
    this property is returned.
    """
    if pvalue:
        properties = [CIMProperty("P", pvalue)]
    else:
        properties = [CIMProperty("Pbt", value=True),
                      CIMProperty("Pbf", value=False),
                      CIMProperty("Puint32", Uint32(4294967295)),
                      CIMProperty("Psint32", Sint32(-2147483648)),
                      CIMProperty("Pint64", Uint64(9999)),
                      CIMProperty("Pdt", DATETIME1_OBJ),
                      CIMProperty("Pstr1", u"Test String")]
    inst = CIMInstance("CIM_Foo", properties)
    return inst


def string_instance(tst_str):
    """
    Build a CIM instance with a single property.
    """
    properties = [CIMProperty("Pstr1", tst_str)]
    inst = CIMInstance("CIM_Foo", properties)
    return inst


# Testcases for format_inst_to_table()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * args: CIMInstance(s) object to be tested and col_width field
    #   * kwargs: Dict of input args for tocimxml().
    #   * exp_xml_str: Expected CIM-XML string, as a tuple/list of parts.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_rtn: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

TESTCASES_FMT_INSTANCE_AS_ROWS = [
    (
        "Verify simple instance to table",
        dict(
            args=([simple_instance()], None),
            kwargs=dict(),
            exp_rtn=[
                ["true", "false", "99", "9999", DATETIME1_STR,
                 u'"Test String"', ]],
        ),
        None, None, True, ),

    (
        "Verify simple instance to table with col limit",
        dict(
            args=([simple_instance()], 30),
            kwargs=dict(),
            exp_rtn=[
                ["true", "false", "99", "9999", DATETIME1_STR,
                 u'"Test String"']],
        ),
        None, None, True, ),

    (
        "Verify simple instance with one string all components overflow line",
        dict(
            args=([simple_instance(pvalue="A B C D")], 4),
            kwargs=dict(),
            exp_rtn=[
                ['"A "\n"B "\n"C "\n"D"']],
        ),
        None, None, True, ),

    (
        "Verify simple instance with one string all components overflow line",
        dict(
            args=([simple_instance(pvalue="ABCD")], 4),
            kwargs=dict(),
            exp_rtn=[
                ['\n"AB"\n"CD"']],
        ),
        None, None, True, ),

    (
        "Verify simple instance with one string overflows line",
        dict(
            args=([simple_instance(pvalue="A B C D")], 8),
            kwargs=dict(),
            exp_rtn=[
                ['"A B C "\n"D"']],
        ),
        None, None, True, ),

    (
        "Verify simple instance withone unit32 max val",
        dict(
            args=([simple_instance(pvalue=Uint32(4294967295))], 8),
            kwargs=dict(),
            exp_rtn=[
                ['4294967295']],
        ),
        None, None, True, ),


    (
        "Verify simple instance with one string fits on line",
        dict(
            args=([simple_instance(pvalue="A B C D")], 12),
            kwargs=dict(),
            exp_rtn=[
                ['"A B C D"']],
        ),
        None, None, True, ),

    (
        "Verify datetime property",
        dict(
            args=([simple_instance(pvalue=DATETIME1_OBJ)], 20),
            kwargs=dict(),
            exp_rtn=[
                ['\n"20140922104920.524"\n"789+000"']],
        ),
        None, None, True, ),

    (
        "Verify datetime property",
        dict(
            args=([simple_instance(pvalue=DATETIME1_OBJ)], 30),
            kwargs=dict(),
            exp_rtn=[
                ['"20140922104920.524789+000"']],
        ),
        None, None, True, ),

    (
        "Verify integer property where len too small",
        dict(
            args=([simple_instance(pvalue=Uint32(999999))], 4),
            kwargs=dict(),
            exp_rtn=[['999999']],
        ),
        None, None, True, ),

    (
        "Verify char16 property",
        dict(
            args=([CIMInstance('P', [CIMProperty('P',
                                                 type='char16',
                                                 value='f')])], 4),
            kwargs=dict(),
            exp_rtn=[[u"'f'"]],
        ),
        None, None, True, ),

    (
        "Verify properties with no value",
        dict(
            args=([CIMInstance('P', [CIMProperty('P', value=None,
                                                 type='char16'),
                                     CIMProperty('Q', value=None,
                                                 type='uint32'),
                                     CIMProperty('R', value=None,
                                                 type='string'), ])], 4),
            kwargs=dict(),
            exp_rtn=[[u'', u'', u'']],
        ),
        None, None, True, ),

    (
        "Verify format of instance with reference property as row entry",
        dict(
            args=([CIMInstance("TST_REFPROP",
                               [CIMProperty(
                                   'P',
                                   type='reference',
                                   reference_class="blah",
                                   value=CIMInstanceName(
                                       "REF_CLN",
                                       keybindings=OrderedDict(k1='v1')))])],
                  30),
            kwargs=dict(),
            exp_rtn=[
                [u'"/:REF_CLN.k1=\\"v1\\""']],
        ),
        None, None, True, ),

]

# TODO: See line 973. We have some test duplication.


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_FMT_INSTANCE_AS_ROWS)
@simplified_test_function
def test_format_insts_as_rows(testcase, args, kwargs, exp_rtn):
    """
    Test the output of the common:format_insts_as_table function
    """

    # The code to be tested
    act_rtn = _format_instances_as_rows(*args, **kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None
    # result is list of lists.  we want to test each item in inner list

    assert len(act_rtn) == len(exp_rtn), \
        "Unexpected number of lines in test desc: {}:\n" \
        "Expected line cnt={}:\n" \
        "{}\n\n" \
        "Actual line cnt={}:\n" \
        "{}\n". \
        format(testcase.desc, len(act_rtn), '\n'.join(act_rtn),
               len(exp_rtn), '\n'.join(exp_rtn))

    assert exp_rtn == act_rtn, \
        "Unequal values for test desc: {}:\n" \
        "Expected = {}:\n" \
        "Actual   = {}:\n". \
        format(testcase.desc, exp_rtn, act_rtn)


# Testcases for format_inst_to_table()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * args: CIMInstance(s) object to be tested, col_width field, ouput_fmt
    #   * kwargs: Dict of input args for tocimxml().
    #   * exp_tbl: Expected Table to be output.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

TESTCASES_PRINT_INSTANCE_AS_TABLE = [
    (
        "Verify print of simple instance to table",
        dict(
            args=([simple_instance()], None, 'simple'),
            kwargs=dict(),
            exp_tbl=[
                'Pbt    Pbf      Pint32    Pint64  Pdt                    '
                'Pstr1\n'
                '-----  -----  --------  --------  ---------------------  '
                '-------------\n'
                'true   false        99      9999  "20140922104920.5247"'
                '  "Test String"\n'
                '                                  "89+000"\n'
            ],
        ),
        None, None, True, ),

    (
        "Verify print of simple instance to table with col limit",
        dict(
            args=([simple_instance2()], 80, 'simple'),
            kwargs=dict(),
            exp_tbl=[
                ["true", "false", "99", '"Test String"']],
        ),
        None, None, True, ),
    (
        "Verify print of instance with reference property",
        dict(
            args=([CIMInstance("CIM_Foo",
                               [CIMProperty(
                                   'P',
                                   type='reference',
                                   reference_class="blah",
                                   value=CIMInstanceName(
                                       "REF_CLN",
                                       keybindings=OrderedDict(k1='v1',
                                                               k2=32)))])],
                  80, 'simple'),
            kwargs=dict(),
            exp_tbl=[
                ["/:REF_CLN.k2=32,k1=\"v1\""]],
        ),
        None, None, True, ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_PRINT_INSTANCE_AS_TABLE)
@simplified_test_function
def test_print_insts_as_table(testcase, args, kwargs, exp_tbl):
    """
    Test the output of the print_insts_as_table function. This primarily
    tests for overall format and the ability of the function to output to
    stdout. The previous test tests the row formatting and handling of
    multiple instances.
    """
    # TODO fix simplified_test_function so we can use so we capture output.
    # capsys in a builtin fixture that must be passed to this function.
    # Currently the simplified_test_function does not allow any other
    # parameters so we cannot use pytest capsys

    # The code to be tested
    _print_instances_as_table(*args, **kwargs)
    # stdout, stderr = capsys.readouterr()
    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # assertexp_tbl, stdout, testcase.desc)


# TODO Test compare and failure in compare_obj

# TODO Test compare with errors

# NOTE: Format table is in test_tableformat.py



#!/usr/bin/env python

"""
This tool can capture the help outputs from a click application and output
the result either as text or in restructured text format.

It executes the click script --help and recursively scrapes all of the
commands from the output, generating an output that is the help text for
every group/command in the script.

All output is to stdout.

It will generate either restructured text or pure text output depending on
the variable USE_RST.

the rst output set a section name for each help subject.

This should be usable with different click generated apps by simply changing
the variable  SCRIPT_NAME to the name of the target scripere t.

There are no executeion inputs since the primary use is to generate information
for review and documentation in a fixed environmet.  The two variables:

SCRIPT_NAME - Name of the click script that will be executed to generate
help information

USE_RST - Boolean. If true, generates .rst output. Otherwise it generates
pure formatted text.

"""


from __future__ import print_function, absolute_import

import sys
import subprocess
try:
    import textwrap
    textwrap.indent  # pylint: disable=pointless-statement
except AttributeError:  # undefined function (wasn't added until Python 3.3)
    def indent(text, amount, ch=' '):  # pylint: disable=invalid-name
        """Indent locally define"""
        padding = amount * ch
        return ''.join(padding + line for line in text.splitlines(True))
else:
    def indent(text, amount, ch=' '):  # pylint: disable=invalid-name
        """Wrap textwrap function"""
        return textwrap.indent(text, amount * ch)

import six

# Flag that allows displaying the data as pure text rather than markdown
# format
USE_RST = True
SCRIPT_NAME = 'pywbemcli'

# SCRIPT_CMD = SCRIPT_NAME  # TODO #103: Reactivate once pywbemcli works
# on Windows
SCRIPT_CMD = 'python -c "import sys; from ' \
    'pywbemtools.pywbemcli.pywbemcli import cli; ' \
    'sys.argv[0]=\'pywbemcli\'; sys.exit(cli())"'

ERRORS = 0
VERBOSE = False


def rst_headline(title, level):
    """
    Format a markdown header line based on the level argument. The rst format
    for headings is generally of the form
    ====================
    title for this thing
    ====================
    """
    level_char_list = ['#', '*', '=', '-', '^', '"']
    try:
        level_char = level_char_list[level]
    except IndexError:
        level_char = '='

    # output anchor in form .. _`smicli commands`:
    anchor = '.. _`%s`:' % title
    title_marker = level_char * len(title)
    if level == 0:
        return '\n%s\n\n%s\n%s\n%s\n' % (anchor, title_marker, title,
                                         title_marker)

    return '\n%s\n\n%s\n%s\n' % (anchor, title, title_marker)


def print_rst_verbatum_text(text_str):
    """
    Print the text on input proceeded by the rst literal block indicator
    """
    print('::\n')
    # indent text for rst. rst requires that block monospace test be
    # indented and preceeded by line with just '::' and followed by
    # empty line. Indent all lines with text

    lines = text_str.split('\n')
    new_lines = []
    for line in lines:
        if line:
            new_lines.append(indent(line, 4))
        else:
            new_lines.append(line)
    print('%s' % '\n'.join(new_lines))


HELP_DICT = {}


def cmd_exists(cmd):
    """
    Determine if the command defined by cmd can be executed in a shell.

    Returns a tuple (rc, msg), where rc==0 indicates that the command can be
    executed, and otherwise rc is the command (or shell) return code and
    msg is an error message.
    """

    if True:  # TODO #103: Debug PATH for pywbemcli not found issue on Windows
        if sys.platform == 'win32':
            echo_cmd = 'echo %PATH%'
        else:
            echo_cmd = 'echo $PATH'
        proc = subprocess.Popen(echo_cmd, shell=True, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        out, _ = proc.communicate()
        print("Debug: %s: %s" % (echo_cmd, out), file=sys.stderr)

    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    out, _ = proc.communicate()
    rc = proc.returncode
    if rc == 0:
        msg = None
    else:
        msg = out.strip()
    return rc, msg


def get_subcmd_group_names(script_cmd, script_name, cmd):
    """
    Execute the script with defined command and help and get the
    groups defined for that help.

    returns list of command groups/commands
    """
    command = '%s %s --help' % (script_cmd, cmd)
    # Disable python warnings for script call.
    if sys.platform != 'win32':
        command = 'export PYTHONWARNINGS="" && %s' % command
    if VERBOSE:
        print('command %s' % command)
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    std_out, std_err = proc.communicate()
    exitcode = proc.returncode

    HELP_DICT[cmd] = std_out

    if six.PY3:
        std_out = std_out.decode()
        std_err = std_err.decode()

    if exitcode:
        raise RuntimeError("Error: Shell execution of command %r failed with "
                           "rc=%s: %s" % (command, exitcode, std_err))
    if std_err:
        raise RuntimeError("Error: Unexpected stderr from command %r:\n"
                           "%s" % (command, std_err))

    # Split stdout into list of lines
    lines = std_out.split('\n')

    # get first word of all lines after line containing 'Commands:'
    group_list = []
    group_state = False
    for line in lines:
        if group_state and line:
            # split line into list of words and get first word as command
            words = line.split()
            group_list.append(words[0])

        # test for line that matchs the word Commands
        if line.strip() == 'Commands:':
            group_state = True
    return group_list


def get_subgroup_names(group_name, script_cmd, script_name):
    """
    Get all the commands for the help_group_name defined on input.
    Executes script and extracts groups after line with 'Commands'
    """
    subcmds_list = get_subcmd_group_names(script_cmd, script_name, group_name)
    space = ' ' if group_name else ''

    return ['%s%s%s' % (group_name, space, name) for name in subcmds_list]


def create_help_cmd_list(script_cmd, script_name):
    """
    Create the command list.
    """
    # Result list of assembled help subcmds
    help_groups_result = []
    # start with empty group, the top level (i.e. pywbemcli --help).
    # This is list of names to process and is extended as we process
    # each group.
    group_names = [""]

    help_groups_result.extend(group_names)
    for name in group_names:
        return_cmds = get_subgroup_names(name, script_cmd, script_name)
        help_groups_result.extend(return_cmds)
        # extend input list with returned assembled groups
        group_names.extend(return_cmds)
    # sort to match order of click
    help_groups_result.sort()
    if USE_RST:
        print(rst_headline("%s Help Command Details" % script_name, 2))
        print('\nThis section shows the help text for each %s '
              'command group and command.\n' % script_name)

    for name in help_groups_result:
        command = '%s %s' % (script_name, name)
        command_help = '%s %s --help' % (script_name, name)
        out = HELP_DICT[name]
        if USE_RST:
            level = len(name.split())  # 0: top, 1: group, 2: command
            if level == 0:
                line = 'Help text for ``%s``:' % script_name
            elif level == 1 and name not in ('repl', 'help'):
                line = 'Help text for ``%s`` (see :ref:`%s command group`):' % \
                    (command, name)
            else:
                line = 'Help text for ``%s`` (see :ref:`%s command`):' % \
                    (command, name)
            if level > 0:
                # Don't generate section header for top level
                print(rst_headline(command_help, 2 + level))
            print('\n\n%s\n\n' % line)
            print_rst_verbatum_text(out.decode())
        else:
            print('%s\n%s command: %s' %
                  (('=' * 50), script_name, command_help))
            print(out.decode())

    return help_groups_result


if __name__ == '__main__':
    # Verify that the script exists. Executing with --help loads click
    # script generates help and exits.
    check_cmd = '%s --help' % SCRIPT_CMD
    rc, msg = cmd_exists(check_cmd)
    if rc != 0:
        print("Error: Shell execution of %r returns rc=%s: %s" %
              (check_cmd, rc, msg), file=sys.stderr)
        sys.exit(1)
    create_help_cmd_list(SCRIPT_CMD, SCRIPT_NAME)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test the pywbemcli general options.  That is those options that are defined and
used in pywbemcli.py and show up in the help output from pywbemcli --help.

NOTE: The --log options are tested in a separate file.
"""
from __future__ import absolute_import, print_function

import os
import sys
import pytest

from .cli_test_extensions import CLITestsBase

SCRIPT_DIR = os.path.dirname(__file__)

SIMPLE_MOCK_FILE_PATH = os.path.join(SCRIPT_DIR, 'simple_mock_model.mof')
PYTHON_MOCK_FILE_PATH = os.path.join(SCRIPT_DIR, 'simple_python_mock_script.py')
BAD_MOF_FILE_PATH = os.path.join(SCRIPT_DIR, 'mof_with_error.mof')
BAD_PY_FILE_PATH = os.path.join(SCRIPT_DIR, 'py_with_error.py')


GENERAL_HELP = """
Usage: pywbemcli [GENERAL-OPTIONS] COMMAND [ARGS]...

  Pywbemcli is a command line WBEM client that uses the DMTF CIM-XML
  protocol to communicate with WBEM servers. Pywbemcli can:

  * Manage the information in WBEM servers CIM objects using the
    operations defined in the DMTF specification.  It can manage CIM
    classes, CIM instances and CIM qualifier declarations in the WBEM
    Server and execute CIM methods and queries on the server.

  * Inspect WBEM server characteristics including namespaces, registered
    profiles, and other server information.

  * Capture detailed information on communication with the WBEM
    server including time statistics and logs of the operations.

  * Maintain a persistent list of named connections to WBEM servers
    and execute operations on them by name.

  Pywbemcli implements command groups and commands to execute the CIM-XML
  operations defined by the DMTF CIM Operations Over HTTP specification
  (DSP0200).

  The general options shown below can also be specified on any of the
  commands, positioned right after the 'pywbemcli' command name.

  For more detailed documentation, see:

      https://pywbemtools.readthedocs.io/en/stable/

Options:
  -n, --name NAME                 Use the WBEM server defined by the WBEM
                                  connection definition NAME. This option is
                                  mutually exclusive with the --server and
                                  --mock-server options, since each defines a
                                  WBEM server. Default: EnvVar PYWBEMCLI_NAME,
                                  or none.
  -m, --mock-server FILE          Use a mock WBEM server that is automatically
                                  created in pywbemcli and populated with CIM
                                  objects that are defined in the specified
                                  MOF file or Python script file. See the
                                  pywbemcli documentation for more
                                  information. This option may be specified
                                  multiple times, and is mutually exclusive
                                  with the --server and --name options, since
                                  each defines a WBEM server. Default: EnvVar
                                  PYWBEMCLI_MOCK_SERVER, or none.
  -s, --server URL                Use the WBEM server at the specified URL
                                  with format: [SCHEME://]HOST[:PORT]. SCHEME
                                  must be "https" (default) or "http". HOST is
                                  a short or long hostname or literal IPV4/v6
                                  address. PORT defaults to 5989 for https and
                                  5988 for http. This option is mutually
                                  exclusive with the --mock-server and --name
                                  options, since each defines a WBEM server.
                                  Default: EnvVar PYWBEMCLI_SERVER, or none.
  -u, --user TEXT                 User name for the WBEM server. Default:
                                  EnvVar PYWBEMCLI_USER, or none.
  -p, --password TEXT             Password for the WBEM server. Default:
                                  EnvVar PYWBEMCLI_PASSWORD, or prompted for
                                  if --user specified.
  --verify / --no-verify          If --verify, client verifies the X.509
                                  server certificate presented by the WBEM
                                  server during TLS/SSL handshake. If --no-
                                  verify client bypasses verification.
                                  Default: EnvVar PYWBEMCLI_VERIFY, or "--
                                  verify".
  --ca-certs FILE                 Path name of a file or directory containing
                                  certificates that will be matched against
                                  the server certificate presented by the WBEM
                                  server during TLS/SSL handshake. Default:
                                  EnvVar PYWBEMCLI_CA_CERTS, or [/etc/pki/ca-
                                  trust/extracted/openssl/ca-bundle.trust.crt,
                                  /etc/ssl/certs, /etc/ssl/certificates].
  -c, --certfile FILE             Path name of a PEM file containing a X.509
                                  client certificate that is used to enable
                                  TLS/SSL 2-way authentication by presenting
                                  the certificate to the WBEM server during
                                  TLS/SSL handshake. Default: EnvVar
                                  PYWBEMCLI_CERTFILE, or none.
  -k, --keyfile FILE              Path name of a PEM file containing a X.509
                                  private key that belongs to the certificate
                                  in the --certfile file. Not required if the
                                  private key is part of the --certfile file.
                                  Default: EnvVar PYWBEMCLI_KEYFILE, or none.
  -t, --timeout INT               Client-side timeout in seconds for
                                  operations with the WBEM server. Default:
                                  EnvVar PYWBEMCLI_TIMEOUT, or 30.
  -U, --use-pull [yes|no|either]  Determines whether pull operations are used
                                  for operations with the WBEM server that
                                  return lists of instances, as follows: "yes"
                                  uses pull operations and fails if not
                                  supported by the server; "no" uses
                                  traditional operations; "either" (default)
                                  uses pull operations if supported by the
                                  server, and otherwise traditional
                                  operations. Default: EnvVar
                                  PYWBEMCLI_USE_PULL, or "either".
  --pull-max-cnt INT              Maximum number of instances to be returned
                                  by the WBEM server in each open or pull
                                  response, if pull operations are used. This
                                  is a tuning parameter that does not affect
                                  the external behavior of the commands.
                                  Default: EnvVar PYWBEMCLI_PULL_MAX_CNT, or
                                  1000
  -T, --timestats                 Show time statistics of WBEM server
                                  operations.
  -d, --default-namespace NAMESPACE
                                  Default namespace, to be used when commands
                                  do not specify the --namespace command
                                  option. Default: EnvVar
                                  PYWBEMCLI_DEFAULT_NAMESPACE, or root/cimv2.
  -o, --output-format FORMAT      Output format for the command result. The
                                  specified format may be overriden since not
                                  all formats apply to all result data types.
                                  FORMAT is a table format
                                  [table|plain|simple|grid|psql|rst|html] or
                                  object format [mof|xml|repr|txt]. Default:
                                  simple.
  -l, --log COMP[=DEST[:DETAIL]],...
                                  Enable logging of the WBEM operations,
                                  defined by a list of log configuration
                                  strings with: COMP: [api|http|all]; DEST:
                                  [file|stderr], default: file; DETAIL:
                                  [all|paths|summary], default: all. Default:
                                  EnvVar PYWBEMCLI_LOG, or all.
  -v, --verbose / --no-verbose    Display extra information about the
                                  processing.
  --version                       Show the version of this command and the
                                  pywbem package and exit.
  -h, --help                      Show this message and exit.

Commands:
  class       Command group for CIM classes.
  connection  Command group for WBEM connection definitions.
  help        Show help message for interactive mode.
  instance    Command group for CIM instances.
  qualifier   Command group for CIM qualifier declarations.
  repl        Enter interactive mode (default).
  server      Command group for WBEM servers.

"""


OK = True
RUN = True
FAIL = False
SKIP = False

TEST_CASES = [
    # desc - Description of test
    # inputs - String, or list of args or dict of 'env', 'args', 'general',
    #          and 'stdin'. See See CLITestsBase.subcmd_test()  for
    #          detailed documentation. This test processor includes an
    #          additional key, `subcmd`
    # exp_response - Dictionary of expected responses (stdout, stderr, rc) and
    #                test definition (test: <testname>).
    #                See CLITestsBase.subcmd_test() for detailed documentation.
    # mock - None or name of files (mof or .py),
    # condition - If True, the test is executed,  Otherwise it is skipped.

    ['Verify -help response.',
     {'general': ['--help'],
      'cmdgrp': 'class',
      'args': ['get', 'blah']},
     {'stdout': GENERAL_HELP,
      'rc': 0,
      'test': 'linesnows'},
     None, OK],

    ['Verify -h response.',
     {'general': ['--help'],
      'cmdgrp': 'class',
      'args': ['get', 'blah']},
     {'stdout': GENERAL_HELP,
      'rc': 0,
      'test': 'linesnows'},
     None, OK],

    ['Verify invalid server definition.',
     {'general': ['-s', 'httpx://blah'],
      'cmdgrp': 'class',
      'args': ['get', 'blah']},
     {'stderr': ['Error: Invalid scheme on server argument. httpx://blah Use '
                 '"http" or "https"'],
      'rc': 1,
      'test': 'in'},
     None, OK],

    ['Verify invalid server port definition fails.',
     {'general': ['-s', 'http://blah:abcd'],
      'cmdgrp': 'class',
      'args': ['get', 'blah']},
     {'stderr': ['Error:', 'ConnectionError', 'Socket error'],
      'rc': 1,
      'test': 'regex'},
     None, OK],

    ['Verify valid --use-pull option parameter yes.',
     {'general': ['-s', 'http://blah', '--use-pull', 'yes'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stdout': ['use-pull: True'],
      'rc': 0,
      'test': 'in'},
     None, SKIP],

    ['Verify valid --use-pull option parameter no.',
     {'general': ['-s', 'http://blah', '--use-pull', 'no'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stdout': ['use-pull: False'],
      'rc': 0,
      'test': 'in'},
     None, SKIP],

    ['Verify valid --use-pull option parameter.',
     {'general': ['-s', 'http://blah', '--use-pull', 'either'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stdout': ['use-pull: None'],
      'rc': 0,
      'test': 'in'},
     None, SKIP],

    ['Verify invalid --use-pull option parameter fails.',
     {'general': ['-s', 'http://blah', '--use-pull', 'blah'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Invalid value for "-U" / "--use-pull": invalid choice: '
                 'blah. (choose from yes, no, either)'],
      'rc': 2,
      'test': 'in'},
     None, SKIP],

    ['Verify valid --pull-max-cnt option parameter.',
     {'general': ['-s', 'http://blah', '--pull-max-cnt', '2000'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stdout': ['pull-max-cnt: 2000'],
      'rc': 0,
      'test': 'in'},
     None, SKIP],

    ['Verify invalid --pull-max-cnt option parameter fails.',
     {'general': ['-s', 'http://blah', '--pull-max-cnt', 'blah'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Error: Invalid value for "--pull-max-cnt": blah is not a '
                 'valid integer'],
      'rc': 2,
      'test': 'in'},
     None, SKIP],

    ['Verify --version general option.',
     {'general': ['-s', 'http://blah', '--version'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stdout': [r'^pywbemcli, version [0-9]+\.[0-9]+\.[0-9]+',
                 r'^pywbem, version [0-9]+\.[0-9]+\.[0-9]+'],
      'rc': 0,
      'test': 'regex'},
     None, OK],

    #
    #  Test --verify and --no-verify general option using the connection show
    #
    ['Verify --verify general option.',
     {'general': ['--server', 'http://blah', '--verify'],
      'args': ['show'],
      'cmdgrp': 'connection'},
     {'stdout': "verify: True",
      'test': 'innows',
      'file': {'before': 'none', 'after': 'None'}},
     None, OK],

    ['Verify --no-verify general option.',
     {'general': ['--server', 'http://blah', '--no-verify'],
      'args': ['show'],
      'cmdgrp': 'connection'},
     {'stdout': "verify: False",
      'test': 'innows',
      'file': {'before': 'none', 'after': 'None'}},
     None, OK],

    ['Verify --verify general options  Default value.',
     {'general': ['--server', 'http://blah'],
      'args': ['show'],
      'cmdgrp': 'connection'},
     {'stdout': "verify: True",
      'test': 'innows',
      'file': {'before': 'none', 'after': 'None'}},
     None, OK],

    ['Verify -m option one file',
     {'general': ['-m', SIMPLE_MOCK_FILE_PATH],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stdout': ['name: not-saved',
                 'server: None',
                 'mock-server:',
                 'simple_mock_model.mof',
                 'default-namespace: root/cimv2'],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify -m option, multiple files',
     {'general': ['-m', SIMPLE_MOCK_FILE_PATH,
                  '-m', PYTHON_MOCK_FILE_PATH],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stdout': ['name', 'default',
                 'server', 'None',
                 'mock-server',
                 'simple_mock_model.mof',
                 'simple_python_mock_script.py',
                 'default-namespace: root/cimv2'],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify --mock-server option, file does not exist',
     {'general': ['--mock-server', 'invalidfilename.mof'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Error: Mock file ',
                 'invalidfilename.mof',
                 ' does not exist'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    ['Verify --mock-server option, file with bad extension',
     {'general': ['--mock-server', 'invalidfilename.mofx'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Mock file ',
                 'invalidfilename.mofx',
                 ' has invalid suffix'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    ['Verify -m option, file with mof containing syntax error',
     {'general': ['-m', BAD_MOF_FILE_PATH],
      'cmdgrp': 'class',
      'args': ['enumerate']},
     {'stderr': ['Mock MOF file', BAD_MOF_FILE_PATH, 'failed compiling',
                 'Aborted!'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    ['Verify -m option, file with python containing syntax error',
     {'general': ['-m', BAD_PY_FILE_PATH],
      'cmdgrp': 'class',
      'args': ['enumerate']},
     {'stderr': [r'Traceback \(most recent call last\)',
                 r'pywbemtools', r'_pywbemcli_operations\.py',
                 r'line ', 'in build_repository',
                 r"NameError: name 'globalsx' is not defined",
                 'Aborted'],
      'rc': 1,
      'test': 'regex'},
     None, OK],

    ['Verify -n option with new name but not in repo fails',
     {'general': ['-n', 'fred'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Error: Connection definition', 'fred',
                 'not found in connections file'],
      'rc': 1,
      'test': 'in'},
     None, OK],

    ['Verify --name option with new name but not in repo fails',
     {'general': ['--name', 'fred'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Error: Connection definition', 'fred',
                 'not found in connections file'],
      'rc': 1,
      'test': 'in'},
     None, OK],

    ['Verify simultaneous --mock-server and --server options fail',
     {'general': ['--mock-server', SIMPLE_MOCK_FILE_PATH,
                  '--server', 'http://localhost'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Error: Conflicting server definitions:',
                 'mock-server:', SIMPLE_MOCK_FILE_PATH,
                 'server:', 'http://localhost'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    ['Verify --mock-server invalid name',
     {'general': ['--mock-server', 'fred'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Error: Mock file',
                 'fred',
                 'has invalid suffix'],
      'rc': 1,
      'test': 'regex'},
     None, OK],

    ['Verify -m option one file and name fails',
     {'general': ['-m', SIMPLE_MOCK_FILE_PATH, '--name', 'MyConnName'],
      'cmdgrp': 'connection',
      'args': ['show']},
     {'stderr': ['Error: Conflicting server definitions:',
                 'mock-server:', SIMPLE_MOCK_FILE_PATH,
                 'name:', 'MyConnName'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    ['Verify --timestats',
     {'general': ['--mock-server', SIMPLE_MOCK_FILE_PATH, '--timestats'],
      'cmdgrp': 'class',
      'args': ['enumerate']},
     {'stdout': ['class CIM_Foo {',
                 '  Count    Exc    Time    ReqLen    ReplyLen  Operation',
                 'EnumerateClasses'],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify --timestats -T',
     {'general': ['--mock-server', SIMPLE_MOCK_FILE_PATH, '-T'],
      'cmdgrp': 'class',
      'args': ['enumerate']},
     {'stdout': ['class CIM_Foo {',
                 '  Count    Exc    Time    ReqLen    ReplyLen  Operation',
                 'EnumerateClasses'],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify uses pull Operation  with option either',
     {'general': ['--mock-server', SIMPLE_MOCK_FILE_PATH,
                  '--timestats',
                  '--use-pull', 'either'],
      'cmdgrp': 'instance',
      'args': ['enumerate', 'CIM_Foo']},
     {'stdout': ['instance of CIM_Foo {',
                 '  Count    Exc    Time    ReqLen    ReplyLen  Operation',
                 ' OpenEnumerateInstances'],
      'rc': 0,
      'test': 'regex'},
     None, OK],

    ['Verify uses pull Operation with option yes',
     {'general': ['--mock-server', SIMPLE_MOCK_FILE_PATH,
                  '--timestats',
                  '--use-pull', 'yes'],
      'cmdgrp': 'instance',
      'args': ['enumerate', 'CIM_Foo']},
     {'stdout': ['instance of CIM_Foo {',
                 '  Count    Exc    Time    ReqLen    ReplyLen  Operation',
                 ' OpenEnumerateInstances'],
      'rc': 0,
      'test': 'regex'},
     None, OK],

    ['Verify uses pull Operation with option no',
     {'general': ['--mock-server', SIMPLE_MOCK_FILE_PATH,
                  '--timestats',
                  '--use-pull', 'no'],
      'cmdgrp': 'instance',
      'args': ['enumerate', 'CIM_Foo']},
     {'stdout': ['instance of CIM_Foo {',
                 '  Count    Exc    Time    ReqLen    ReplyLen  Operation',
                 ' EnumerateInstances'],
      'rc': 0,
      'test': 'regex'},
     None, OK],

    ['Verify --mock-server and -server not allowed',
     {'general': ['--mock-server', SIMPLE_MOCK_FILE_PATH,
                  '--server', 'http://blah'],
      'cmdgrp': 'instance',
      'args': ['enumerate', 'CIM_Foo']},
     {'stderr': ['Error: Conflicting server definitions:',
                 'mock-server:', SIMPLE_MOCK_FILE_PATH,
                 'server:', 'http://blah'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    #
    #   The following is a new sequence but depends on the repo being empty
    #   It creates, shows and deletes a single server definition to demo
    #   effect of other parameters than the --name parameter
    #

    ['Verify Create a connection for test.',
     {'general': ['--server', 'http://blah',
                  '--timeout', '45',
                  '--default-namespace', 'root/blah',
                  '--user', 'john',
                  '--password', 'pw',
                  '--no-verify',
                  '--certfile', 'mycertfile.pem',
                  '--keyfile', 'mykeyfile.pem'],
      'args': ['save', 'generaltest1'],
      'cmdgrp': 'connection', },
     {'stdout': "",
      'test': 'innows'},
     None, OK],

    ['Verify --name and other options fails',
     {'general': ['--name', 'generaltest1', '--timeout', '90'],
      'cmdgrp': 'connection',
      'args': ['delete', 'generaltest1']},
     {'stderr': ['timeout 90', 'invalid when'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    ['Verify connection gets deleted.',
     {'cmdgrp': 'connection',
      'args': ['delete', 'generaltest1']},
     {'stdout': '',
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify connection already deleted.',
     {'cmdgrp': 'connection',
      'args': ['delete', 'generaltest1']},
     {'stderr': ['Connection repository', 'empty'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    ['Verify connection without server definition and command that '
     ' requires connection fails.',
     {'cmdgrp': 'class',
      'args': ['enumerate']},
     {'stderr': ['No server specified for a command that requires a WBEM '
                 'server. To specify a server, use the "--server", '
                 '"--mock-server", or "--name" general options, or set the '
                 'corresponding environment variables, or in interactive mode '
                 'use "connection select"'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    ['Verify --keyfile allowed only if --certfile exists.',
     {'cmdgrp': 'class',
      'args': ['enumerate'],
      'general': ['--keyfile', 'mykey.pem']},
     {'stderr': ['The --keyfile option',
                 'is allowed only if the --certfile option is also used'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    #
    #  Test use of default connection
    #

    ['Verify Create a connection to test default connection.',
     {'general': ['--server', 'http://blah'],
      'args': ['save', 'test-default'],
      'cmdgrp': 'connection', },
     {'stdout': '',
      'test': 'innows'},
     None, OK],

    ['Verify select of test-default.',
     {'args': ['select', 'test-default'],
      'cmdgrp': 'connection', },
     {'stdout': ['test-default', 'current'],
      'test': 'innows'},
     None, OK],

    ['Verify shows test-default connection.',
     {'args': ['show', 'test-default'],
      'cmdgrp': 'connection', },
     {'stdout': ['test-default'],
      'test': 'innows'},
     None, OK],

    ['Verify select of test-default connection.',
     {'args': ['select', 'test-default', '--default'],
      'cmdgrp': 'connection', },
     {'stdout': ['test-default', 'current'],
      'test': 'innows'},
     None, OK],

    ['Verify show current which is test-default',
     {'args': ['show'],
      'cmdgrp': 'connection', },
     {'stdout': ['test-default'],
      'test': 'innows'},
     None, OK],

    ['Delete test-default.',
     {'args': ['delete', 'test-default'],
      'cmdgrp': 'connection', },
     {'stdout': "",
      'test': 'innows'},
     None, OK],

    ['Verify delete worked by requesting again expecting failure.',
     {'args': ['delete', 'test-default'],
      'cmdgrp': 'connection', },
     {'stderr': ['Connection repository', 'empty'],
      'rc': 1,
      'test': 'innows'},
     None, OK],

    #
    # Test using environment variables as input
    #
    ['Verify setting one env var for input.',
     {'env': {'PYWBEMCLI_SERVER': 'http://blah'},
      'cmdgrp': 'connection',
      'args': ['show', '--show-password']},
     {'stdout': ['name', 'default',
                 'server', 'http://blah',
                 'default-namespace', 'root/cimv2',
                 'user', 'None',
                 'password', 'None',
                 'timeout', '30',
                 'verify', 'True',
                 'certfile', 'None',
                 'keyfile', 'None',
                 'mock-server'],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify setting all env vars for input.',
     {'env': {'PYWBEMCLI_SERVER': 'http://blah',
              'PYWBEMCLI_DEFAULT_NAMESPACE': 'fred/fred',
              'PYWBEMCLI_USER': 'RonaldMcDonald',
              'PYWBEMCLI_PASSWORD': 'abcxdfG',
              'PYWBEMCLI_TIMEOUT': '99',
              'PYWBEMCLI_VERIFY': 'True',
              'PYWBEMCLI_CERTFILE': 'certfile.pem',
              'PYWBEMCLI_KEYFILE': 'keyfile.pem',
              'PYWBEMCLI_USE_PULL': 'no',
              'PYWBEMCLI_PULL_MAX_CNT': '10',
              'PYWBEMCLI_LOG': 'api=all', },
      'cmdgrp': 'connection',
      'args': ['show', '--show-password']},
     {'stdout': ['name', 'default',
                 'server', 'http://blah',
                 'default-namespace', 'fred/fred',
                 'user', 'RonaldMcDonald',
                 'password', 'abcxdfG',
                 'timeout', '99',
                 'verify', 'True',
                 'certfile', 'certfile.pem',
                 'keyfile', 'keyfile.pem'],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify mixed env var and input options.',
     {'env': {'PYWBEMCLI_DEFAULT_NAMESPACE': 'fred/fred',
              'PYWBEMCLI_USER': 'RonaldMcDonald',
              'PYWBEMCLI_PASSWORD': 'abcxdfG',
              'PYWBEMCLI_TIMEOUT': '99',
              'PYWBEMCLI_NO_VERIFY': 'True',
              'PYWBEMCLI_CERTFILE': 'certfile.pem',
              'PYWBEMCLI_KEYFILE': 'keyfile.pem',
              'PYWBEMCLI_USE_PULL': 'no',
              'PYWBEMCLI_PULL_MAX_CNT': '10',
              'PYWBEMCLI_LOG': 'api=all', },
      'cmdgrp': 'connection',
      'general': ['--server', 'http://blah'],
      'args': ['show', '--show-password']},
     {'stdout': ['server', 'http://blah',
                 'default-namespace', 'fred/fred',
                 'user', 'RonaldMcDonald',
                 'password', 'abcxdfG',
                 'timeout', '99',
                 'verify', 'True',
                 'certfile', 'certfile.pem',
                 'keyfile', 'keyfile.pem'],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    #
    #  Test setting general options in interactive mode
    #

    ['Verify multiple commands in interacive mode.',
     {'general': ['--server', 'http://blah', ],
      # args not allowed in interactive mode
      'stdin': ['--certfile cert1.pem --keyfile keys1.pem connection show',
                'connection show',
                '--certfile cert2.pem --keyfile keys2.pem connection show',
                'connection show'],             # should show None
      'cmdgrp': 'connection', },
     {'stdout': ['certfile: None',              # at startup and other times
                 'certfile: cert1.pem',         # after first change
                 'certfile: cert2.pem',         # after second change
                 'keyfile: None',               # at startup and other times
                 'keyfile: keys1.pem',
                 'keyfile: keys2.pem'],
      'test': 'innows'},
     None, OK],

    ['Verify Change --mock-server to --server in interactive mode.',
     {'general': ['--mock-server', SIMPLE_MOCK_FILE_PATH, ],
      # args not allowed in interactive mode
      'stdin': ['connection show',
                '--server  http://blah connection show',
                '--mock-server tests/unit/simple_mock_model.mof '
                'connection show',
                'connection show'],
      'cmdgrp': None,
      },
     {'stdout': ['server : None',
                 'server: http://blah',
                 'mock-server:',
                 'simple_mock_model.mof'],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    #
    #   The following is a sequence
    #
    ['Verify Create a connection for test of mods through general opts.',
     {'general': ['--server', 'http://blah',
                  '--timeout', '45',
                  '--default-namespace', 'root/fred',
                  '--user', 'RonaldMcDonald',
                  '--password', 'pw',
                  '--no-verify',
                  '--certfile', 'certfile.pem',
                  '--keyfile', 'keyfile.pem'],
      'args': ['save', 'testGeneralOpsMods'],
      'cmdgrp': 'connection', },
     {'stdout': "",
      'test': 'innows'},
     None, OK],

    ['Verify show current which is testGeneralOpsMods',
     {'args': ['show', 'testGeneralOpsMods', '--show-password'],
      'cmdgrp': 'connection', },
     {'stdout': ['testGeneralOpsMods',
                 'server', 'http://blah',
                 'default-namespace', 'root/fred',
                 'user', 'RonaldMcDonald',
                 'password', 'pw',
                 'timeout', '45',
                 'verify', 'False',
                 'certfile', 'certfile.pem',
                 'keyfile', 'keyfile.pem'],
      'test': 'innows'},
     None, OK],

    ['Verify Change all server parameters and show.',
     {'general': ['--name', 'testGeneralOpsMods'],
      # args not allowed in interactive mode
      'stdin': ['--server  http://blahblah --timeout 90 --user Fred '
                '--default-namespace root/john --password  abcd '
                ' --verify --certfile c1.pem --keyfile k1.pem '
                'connection show --show-password'],
      'cmdgrp': None,
      },
     {'stdout': ['testGeneralOpsMods',
                 'server : http://blahblah',
                 'default-namespace', 'root/john',
                 'user', 'Fred',
                 'password', 'abcd',
                 'timeout', '90',
                 'verify', 'True',
                 'certfile', 'c1.pem',
                 'keyfile', 'k1.pem'
                 ],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Change all parameters and save as t1.',
     {'general': ['--name', 'testGeneralOpsMods'],
      # args not allowed in interactive mode
      'stdin': ['--server  http://blahblah --timeout 90 --user Fred '
                '--default-namespace root/john --password  abcd '
                ' --verify --certfile c1.pem --keyfile k1.pem connection '
                'save t1',
                'connection list'],
      'cmdgrp': None,
      },
     {'stdout': [''],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify  load t1 and changed parameters are correct',
     {'general': ['--name', 't1'],
      'args': ['show', '--show-password'],
      'cmdgrp': 'connection'},
     {'stdout': ['t1',
                 'server : http://blahblah',
                 'default-namespace', 'root/john',
                 'user', 'Fred',
                 'password', 'abcd',
                 'timeout', '90',
                 'verify', 'True',
                 'certfile', 'c1.pem',
                 'keyfile', 'k1.pem'
                 ],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Delete testGeneralOpsMods.',
     {'args': ['delete', 'testGeneralOpsMods'],
      'cmdgrp': 'connection', },
     {'stdout': "",
      'test': 'innows'},
     None, OK],

    #
    #  Test verbose option
    #
    ['Verify --verbose on by doing a connection show with --verbose',
     {'general': ['--name', 't1', '--verbose'],
      'args': 'show',
      'cmdgrp': 'connection'},
     {'stdout': ['t1',
                 'server', 'http://blahblah',
                 'default-namespace', 'root/john',
                 'user', 'Fred',
                 'timeout', '90',
                 'verify', 'True',
                 'certfile', 'c1.pem',
                 'keyfile', 'k1.pem',
                 'COMMAND', 'show',
                 'params=', "'name': None",
                 "show_password': False"],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify --verbose on by doing a connection show with -v',
     {'general': ['--name', 't1', '-v'],
      'args': 'show',
      'cmdgrp': 'connection'},
     {'stdout': ['t1',
                 'server', 'http://blahblah',
                 'default-namespace', 'root/john',
                 'user', 'Fred',
                 'timeout', '90',
                 'verify', 'True',
                 'certfile', 'c1.pem',
                 'keyfile', 'k1.pem',
                 'COMMAND', 'show',
                 'params=', "'name': None"],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify --no-verbose',
     {'general': ['--name', 't1', '--no-verbose'],
      'args': ['show', '--show-password'],
      'cmdgrp': 'connection'},
     {'stdout': ['t1',
                 'server : http://blahblah',
                 'default-namespace', 'root/john',
                 'user', 'Fred',
                 'password', 'abcd',
                 'timeout', '90',
                 'verify', 'True',
                 'certfile', 'c1.pem',
                 'keyfile', 'k1.pem'
                 ],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify -verbose but turned off on interactive cmd',
     {'general': ['--name', 't1', '-v'],
      'stdin': ['--no-verbose connection show --show-password'],
      'cmdgrp': None},
     {'stdout': ['t1',
                 'server : http://blahblah',
                 'default-namespace', 'root/john',
                 'user', 'Fred',
                 'password', 'abcd',
                 'timeout', '90',
                 'verify', 'True',
                 'certfile', 'c1.pem',
                 'keyfile', 'k1.pem'
                 ],
      'rc': 0,
      'test': 'innows'},
     None, OK],

    ['Verify -verbose but turned off on interactive cmd',
     {'general': ['--name', 't1', '-v'],
      'stdin': ['--no-verbose connection show'],
      'cmdgrp': None},
     {'stdout': ['COMMAND', 'params='],
      'rc': 0,
      'test': 'not-innows'},
     None, OK],

    ['Delete testGeneralOpsMods.',
     {'args': ['delete', 't1'],
      'cmdgrp': 'connection', },
     {'stdout': "",
      'test': 'innows'},
     None, OK],

    #
    #  Test that bad mock_error file names do not cause Abort but bad MOF
    #  or python script do cause Abort
    #

    ['Verify interactive create mock with bad file name does not fail.',
     {'general': [],
      'stdin': ['--server http://blah --user fred --password fred',
                'connection save connectiontoprovestdincontinues',
                '--mock-server DoesNotExist.mof class enumerate',
                '-m DoesNotExist.py class enumerate',
                'connection select connectiontoprovestdincontinues',
                'connection show',
                'connection delete connectiontoprovestdincontinues']},
     {'stdout': ['name', 'fred',
                 'server', 'http://blah',
                 'default-namespace', 'root/cimv2',
                 'user',
                 'fred',
                 'Deleted connection "connectiontoprovestdincontinues"'],
      'stderr': ['DoesNotExist.mof',
                 'DoesNotExist.py',
                 'does not exist'],
      'test': 'innows'},
     None, OK],

    # TODO this test fails on windows because the statement that builds the
    # --mock-server ... class enumerate inserts the filepath without
    # the backslashes. This is a windows issue, and problem occurs for both
    # python 2 and 3
    ['Verify interactive MOF or PY failures causes Abort.',
     {'general': [],
      'stdin': ['--server http://blah --user fred --password fred ',
                'connection save fred',
                '--mock-server %s class enumerate' % BAD_MOF_FILE_PATH,
                'connection select fred',
                'connection show',
                'connection delete fred'],
      'platform': 'win32'},  # ignore this platform.  See comments above
     {'stdout': ['Deleted connection "fred"'],
      'rc': 1,
      'test': 'not-innows'},
     None, OK],

    # NOTE That we are not using this because of possible issue with previous
    # test and windows.  Enable this before we add any more tests in future.
    ['Delete fred. This one will fail if previous does not abort',
     {'args': ['delete', 'fred'],
      'cmdgrp': 'connection', },
     {'stdout': "",
      'test': 'innows'},
     None, FAIL],

]

# TODO add test for pull operations with pull ops max size variations


class TestGeneralOptions(CLITestsBase):
    """
    Test the general options including statistics,  --server,
    --timeout, --use-pull, --pull-max-cnt, --output-format
    """
    @pytest.mark.parametrize(
        "desc, inputs, exp_response, mock, condition", TEST_CASES)
    def test_execute_pywbemcli(self, desc, inputs, exp_response, mock,
                               condition):
        """


        Execute pybemcli with the defined input and test output.
        """
        # Temp bypass of windows platform because of issue with one test
        if 'platform' in inputs:
            if sys.platform == inputs['platform']:
                return

        cmd_grp = inputs['cmdgrp'] if 'cmdgrp' in inputs else ''

        self.command_test(desc, cmd_grp, inputs, exp_response,
                          mock, condition, verbose=False)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click Command definition for the class command group which includes
cmds for get, enumerate, list of classes.
"""
from __future__ import absolute_import, print_function

import click
from pywbem import Error, CIMError, CIM_ERR_NOT_FOUND
from .pywbemcli import cli
from ._common import display_cim_objects, parse_wbemuri_str, \
    pick_instance, resolve_propertylist, create_ciminstance, \
    filter_namelist, CMD_OPTS_TXT, format_table, verify_operation, \
    process_invokemethod
from ._common_options import add_options, propertylist_option, \
    names_only_option, include_classorigin_instance_option, namespace_option, \
    summary_option, verify_option, multiple_namespaces_option
from .config import DEFAULT_QUERY_LANGUAGE


#
#   Common option definitions for instance group
#

# NOTE: A number of the options use double-dash as the short form.  In those
# cases, a third definition of the options without the double-dash defines
# the corresponding option name, ex. 'include_qualifiers'. It should be
# defined with underscore and not dash


# This is instance-only because the default is False for include-qualifiers
# on instances but True on classes
include_qualifiers_get_option = [              # pylint: disable=invalid-name
    click.option('--iq', '--include-qualifiers', 'include_qualifiers',
                 is_flag=True, required=False,
                 help='Include qualifiers in the returned instance. '
                      'Not all servers return qualifiers on instances. '
                      'Default: Do not include qualifiers.')]

include_qualifiers_list_option = [              # pylint: disable=invalid-name
    click.option('--iq', '--include-qualifiers', 'include_qualifiers',
                 is_flag=True, required=False,
                 help='When traditional operations are used, include '
                      'qualifiers in the returned instances. '
                      'Some servers may ignore this option. '
                      'By default, and when pull operations are used, '
                      'qualifiers will never be included.')]

# specific to instance because DeepInheritance differs between class and
# instance operations.
deep_inheritance_enum_option = [              # pylint: disable=invalid-name
    click.option('--di', '--deep-inheritance', 'deep_inheritance',
                 is_flag=True, required=False,
                 help='Include subclass properties in the returned '
                      'instances. '
                      'Default: Do not include subclass properties.')]

local_only_get_option = [              # pylint: disable=invalid-name
    click.option('--lo', '--local-only', 'local_only', is_flag=True,
                 required=False,
                 help='Do not include superclass properties in the returned '
                      'instance. '
                      'Some servers may ignore this option. '
                      'Default: Include superclass properties.')]

local_only_list_option = [              # pylint: disable=invalid-name
    click.option('--lo', '--local-only', 'local_only', is_flag=True,
                 required=False,
                 help='When traditional operations are used, do not include '
                      'superclass properties in the returned instances. '
                      'Some servers may ignore this option. '
                      'By default, and when pull operations are used, '
                      'superclass properties will always be included.')]

property_create_option = [              # pylint: disable=invalid-name
    click.option('-p', '--property', type=str, metavar='PROPERTYNAME=VALUE',
                 required=False, multiple=True,
                 help='Initial property value for the new instance. '
                      'May be specified multiple times. '
                      'Array property values are specified as a '
                      'comma-separated list; embedded instances are not '
                      'supported. '
                      'Default: No initial properties provided.')]

property_modify_option = [              # pylint: disable=invalid-name
    click.option('-p', '--property', type=str, metavar='PROPERTYNAME=VALUE',
                 required=False, multiple=True,
                 help='Property to be modified, with its new value. '
                      'May be specified once for each property to be '
                      'modified. '
                      'Array property values are specified as a '
                      'comma-separated list; embedded instances are not '
                      'supported. '
                      'Default: No properties modified.')]

filter_query_language_option = [              # pylint: disable=invalid-name
    click.option('--fql', '--filter-query-language', 'filter_query_language',
                 type=str, metavar='QUERY-LANGUAGE', default=None,
                 help='The filter query language to be used with '
                      '--filter-query. '
                      'Default: DMTF:FQL.')]

filter_query_option = [              # pylint: disable=invalid-name
    click.option('--fq', '--filter-query', 'filter_query', type=str,
                 metavar='QUERY-STRING', default=None,
                 help='When pull operations are used, filter the instances in '
                      'the result via a filter query. '
                      'By default, and when traditional operations are used, '
                      'no such filtering takes place.')]


##########################################################################
#
#   Click command group and command definitions
#
###########################################################################


@cli.group('instance', options_metavar=CMD_OPTS_TXT)
def instance_group():
    """
    Command group for CIM instances.

    This command group defines commands to inspect instances, to invoke
    methods on instances, and to create and delete instances.

    Modification of instances is not currently supported.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'instance' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@instance_group.command('get', options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME', required=True)
@add_options(local_only_get_option)
@add_options(include_qualifiers_get_option)
@add_options(include_classorigin_instance_option)
@add_options(propertylist_option)
@add_options(namespace_option)
@click.pass_obj
def instance_get(context, instancename, **options):
    """
    Get an instance of a class.

    The instance can be specified in two ways:

    1. By specifying an untyped WBEM URI of an instance path in the
    INSTANCENAME argument. The CIM namespace in which the instance is looked up
    is the namespace specified in the WBEM URI, or otherwise the namespace
    specified in the --namespace option, or otherwise the default namespace of
    the connection. Any host name in the WBEM URI will be ignored.

    2. By specifying a class name with wildcard for the keys in the
    INSTANCENAME argument, i.e. "CLASSNAME.?". The instances of the specified
    class are displayed and the user is prompted for an index number to select
    an instance. The namespace in which the instances are looked up is the
    namespace specified in the --namespace option, or otherwise the default
    namespace of the connection.

    The --local-only, --include-qualifiers, --include-classorigin, and
    --propertylist options determine which parts are included in the retrieved
    instance.

    In the output, the instance will formatted as defined by the
    --output-format general option.
    """
    context.execute_cmd(lambda: cmd_instance_get(context, instancename,
                                                 options))


@instance_group.command('delete', options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME', required=True)
@add_options(namespace_option)
@click.pass_obj
def instance_delete(context, instancename, **options):
    """
    Delete an instance of a class.

    The CIM instance to be deleted can be specified as follows:

    1. By specifying an untyped WBEM URI of an instance path in the
    INSTANCENAME argument. The CIM namespace in which the instance is looked up
    is the namespace specified in the WBEM URI, or otherwise the namespace
    specified in the --namespace option, or otherwise the default namespace of
    the connection. Any host name in the WBEM URI will be ignored.

    2. By specifying a class name with wildcard for the keys in the
    INSTANCENAME argument, i.e. "CLASSNAME.?". The instances of the specified
    class are displayed and the user is prompted for an index number to select
    an instance. The namespace in which the instances are looked up is the
    namespace specified in the --namespace option, or otherwise the default
    namespace of the connection.
    """
    context.execute_cmd(lambda: cmd_instance_delete(context, instancename,
                                                    options))


@instance_group.command('create', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@add_options(property_create_option)
@add_options(verify_option)
@add_options(namespace_option)
@click.pass_obj
def instance_create(context, classname, **options):
    """
    Create an instance of a class in a namespace.

    Create a CIM instance of the specified creation class (CLASSNAME
    argument) in the specified CIM namespace (--namespace option), with
    the specified properties (--property options) and display the CIM instance
    path of the created instance. If no namespace was specified, the default
    namespace of the connection is used.

    The properties to be initialized and their new values are specified using
    the --property option, which may be specified multiple times.

    Pywbemcli retrieves the class definition from the server in order to
    verify that the specified properties are consistent with the property
    characteristics in the class definition.

    Example:

      pywbemcli instance create CIM_blah -P id=3 -P arr="bla bla",foo
    """
    context.execute_cmd(lambda: cmd_instance_create(context, classname,
                                                    options))


@instance_group.command('modify', options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME', required=True)
@add_options(property_modify_option)
@click.option('--pl', '--propertylist', 'propertylist', multiple=True, type=str,
              default=None, required=False, metavar='PROPERTYLIST',
              help='Reduce the properties to be modified (as per '
              '--property) to a specific property list. '
              'Multiple properties may be specified with either a '
              'comma-separated list or by using the option multiple '
              'times. The empty string will cause no properties to '
              'be modified. '
              'Default: Do not reduce the properties to be modified.')
@add_options(verify_option)
@add_options(namespace_option)
@click.pass_obj
def instance_modify(context, instancename, **options):
    """
    Modify properties of an instance.

    The CIM instance to be modified can be specified in two ways:

    1. By specifying an untyped WBEM URI of an instance path in the
    INSTANCENAME argument. The CIM namespace in which the instance is looked up
    is the namespace specified in the WBEM URI, or otherwise the namespace
    specified in the --namespace option, or otherwise the default namespace of
    the connection. Any host name in the WBEM URI will be ignored.

    2. By specifying a class name with wildcard for the keys in the
    INSTANCENAME argument, i.e. "CLASSNAME.?". The instances of the specified
    class are displayed and the user is prompted for an index number to select
    an instance. The namespace in which the instances are looked up is the
    namespace specified in the --namespace option, or otherwise the default
    namespace of the connection.

    The properties to be modified and their new values are specified using the
    --property option, which may be specified multiple times.

    The --propertylist option allows restricting the set of properties to be
    modified. Given that the set of properties to be modified is already
    determined by the specified --property options, it does not need to be
    specified.

    Example:

      pywbemcli instance modify CIM_blah.fred=3 -P id=3 -P arr="bla bla",foo
    """
    context.execute_cmd(lambda: cmd_instance_modify(context, instancename,
                                                    options))


@instance_group.command('invokemethod', options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME', required=True)
@click.argument('methodname', type=str, metavar='METHODNAME', required=True)
@click.option('-p', '--parameter', type=str, metavar='PARAMETERNAME=VALUE',
              required=False, multiple=True,
              help='Specify a method input parameter with its value. '
                   'May be specified multiple times. '
                   'Array property values are specified as a comma-separated '
                   'list; embedded instances are not supported. '
                   'Default: No input parameters.')
@add_options(namespace_option)
@click.pass_obj
def instance_invokemethod(context, instancename, methodname, **options):
    """
    Invoke a method on an instance.

    Invoke a CIM method (METHODNAME argument) on a CIM instance with the
    specified input parameters (--parameter options), and display the method
    return value and output parameters.

    The CIM instance can be specified in two ways:

    1. By specifying an untyped WBEM URI of an instance path in the
    INSTANCENAME argument. The CIM namespace in which the instance is looked up
    is the namespace specified in the WBEM URI, or otherwise the namespace
    specified in the --namespace option, or otherwise the default namespace of
    the connection. Any host name in the WBEM URI will be ignored.

    2. By specifying a class name with wildcard for the keys in the
    INSTANCENAME argument, i.e. "CLASSNAME.?". The instances of the specified
    class are displayed and the user is prompted for an index number to select
    an instance. The namespace in which the instances are looked up is the
    namespace specified in the --namespace option, or otherwise the default
    namespace of the connection.

    The method input parameters are specified using the --parameter option,
    which may be specified multiple times.

    Pywbemcli retrieves the class definition of the creation class of the
    instance from the server in order to verify that the specified input
    parameters are consistent with the parameter characteristics in the method
    definition.

    Use the 'class invokemethod' command to invoke CIM methods on CIM classes.

    Example:

      pywbemcli -n myconn instance invokemethod CIM_x.id='hi" methodx -p id=3
    """
    context.execute_cmd(lambda: cmd_instance_invokemethod(context,
                                                          instancename,
                                                          methodname,
                                                          options))


@instance_group.command('enumerate', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@add_options(local_only_list_option)
@add_options(deep_inheritance_enum_option)
@add_options(include_qualifiers_list_option)
@add_options(include_classorigin_instance_option)
@add_options(propertylist_option)
@add_options(namespace_option)
@add_options(names_only_option)
@add_options(summary_option)
@add_options(filter_query_option)
@add_options(filter_query_language_option)
@click.pass_obj
def instance_enumerate(context, classname, **options):
    """
    List the instances of a class.

    Enumerate the CIM instances of the specified class (CLASSNAME argument),
    including instances of subclasses in the specified CIM namespace
    (--namespace option), and display the returned instances, or instance paths
    if --names-only was specified. If no namespace was specified, the default
    namespace of the connection is used.

    The instances to be retrieved can be filtered by the --filter-query option.

    The --local-only, --deep-inheritance, --include-qualifiers,
    --include-classorigin, and --propertylist options determine which parts
    are included in each retrieved instance.

    The --names-only option can be used to show only the instance paths.

    In the output, the instances and instance paths will be formatted as
    defined by the --output-format general option. Table formats on instances
    will be replaced with MOF format.
    """
    context.execute_cmd(lambda: cmd_instance_enumerate(context, classname,
                                                       options))


@instance_group.command('references', options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME', required=True)
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help='Filter the result set by result class name. '
                   'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False, metavar='PROPERTYNAME',
              help='Filter the result set by source end role name.')
@add_options(include_qualifiers_list_option)
@add_options(include_classorigin_instance_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(filter_query_option)
@add_options(filter_query_language_option)
@click.pass_obj
def instance_references(context, instancename, **options):
    """
    List the instances referencing an instance.

    List the CIM (association) instances that reference the specified CIM
    instance, and display the returned instances, or instance paths if
    --names-only was specified.

    The CIM instance can be specified in two ways:

    1. By specifying an untyped WBEM URI of an instance path in the
    INSTANCENAME argument. The CIM namespace in which the instance is looked up
    is the namespace specified in the WBEM URI, or otherwise the namespace
    specified in the --namespace option, or otherwise the default namespace of
    the connection. Any host name in the WBEM URI will be ignored.

    2. By specifying a class name with wildcard for the keys in the
    INSTANCENAME argument, i.e. "CLASSNAME.?". The instances of the specified
    class are displayed and the user is prompted for an index number to select
    an instance. The namespace in which the instances are looked up is the
    namespace specified in the --namespace option, or otherwise the default
    namespace of the connection.

    The instances to be retrieved can be filtered by the --filter-query, --role
    and --result-class options.

    The --include-qualifiers, --include-classorigin, and --propertylist options
    determine which parts are included in each retrieved instance.

    The --names-only option can be used to show only the instance paths.

    In the output, the instances and instance paths will be formatted as
    defined by the --output-format general option. Table formats on instances
    will be replaced with MOF format.
    """
    context.execute_cmd(lambda: cmd_instance_references(context, instancename,
                                                        options))


@instance_group.command('associators', options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME', required=True)
@click.option('--ac', '--assoc-class', 'assoc_class', type=str, required=False,
              metavar='CLASSNAME',
              help='Filter the result set by association class name. '
                   'Subclasses of the specified class also match.')
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help='Filter the result set by result class name. '
                   'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help='Filter the result set by source end role name.')
@click.option('--rr', '--result-role', 'result_role', type=str, required=False,
              metavar='PROPERTYNAME',
              help='Filter the result set by far end role name.')
@add_options(include_qualifiers_list_option)
@add_options(include_classorigin_instance_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(filter_query_option)
@add_options(filter_query_language_option)
@click.pass_obj
def instance_associators(context, instancename, **options):
    """
    List the instances associated with an instance.

    List the CIM instances that are associated with the specified CIM instance,
    and display the returned instances, or instance paths if --names-only was
    specified.

    The CIM instance can be specified in two ways:

    1. By specifying an untyped WBEM URI of an instance path in the
    INSTANCENAME argument. The CIM namespace in which the instance is looked up
    is the namespace specified in the WBEM URI, or otherwise the namespace
    specified in the --namespace option, or otherwise the default namespace of
    the connection. Any host name in the WBEM URI will be ignored.

    2. By specifying a class name with wildcard for the keys in the
    INSTANCENAME argument, i.e. "CLASSNAME.?". The instances of the specified
    class are displayed and the user is prompted for an index number to select
    an instance. The namespace in which the instances are looked up is the
    namespace specified in the --namespace option, or otherwise the default
    namespace of the connection.

    The instances to be retrieved can be filtered by the --filter-query,
    --role, --result-role, --assoc-class, and --result-class options.

    The --include-qualifiers, --include-classorigin, and --propertylist options
    determine which parts are included in each retrieved instance.

    The --names-only option can be used to show only the instance paths.

    In the output, the instances and instance paths will be formatted as
    defined by the --output-format general option. Table formats on instances
    will be replaced with MOF format.
    """
    context.execute_cmd(lambda: cmd_instance_associators(context, instancename,
                                                         options))


@instance_group.command('query', options_metavar=CMD_OPTS_TXT)
@click.argument('query', type=str, required=True, metavar='QUERY-STRING')
@click.option('--ql', '--query-language', 'query_language', type=str,
              metavar='QUERY-LANGUAGE', default=DEFAULT_QUERY_LANGUAGE,
              help='The query language to be used with --query. '
              'Default: {default}.'.
              format(default=DEFAULT_QUERY_LANGUAGE))
@add_options(namespace_option)
@add_options(summary_option)
@click.pass_obj
def instance_query(context, query, **options):
    """
    Execute a query on instances in a namespace.

    Execute the specified query (QUERY_STRING argument) in the specified CIM
    namespace (--namespace option), and display the returned instances. If no
    namespace was specified, the default namespace of the connection is used.

    In the output, the instances will formatted as defined by the
    --output-format general option.
    """
    context.execute_cmd(lambda: cmd_instance_query(context, query, options))


@instance_group.command('count', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME-GLOB',
                required=False)
@click.option('-s', '--sort', is_flag=True, required=False,
              help='Sort by instance count. Otherwise sorted by class name.')
@add_options(multiple_namespaces_option)
@click.pass_obj
def instance_count(context, classname, **options):
    """
    Count the instances of each class with matching class name.

    Display the count of instances of each CIM class whose class name
    matches the specified wildcard expression (CLASSNAME-GLOB) in all CIM
    namespaces of the WBEM server, or in the specified namespaces
    (--namespace option).  This differs from instance enumerate, etc. in that
    it counts the instances specifically for the classname of each instance
    returned, not including subclasses.

    The CLASSNAME-GLOB argument is a wildcard expression that is matched on
    class names case insensitively.
    The special characters from Unix file name wildcarding are supported
    ('*' to match zero or more characters, '?' to match a single character,
    and '[]' to match character ranges). To avoid shell expansion of wildcard
    characters, the CLASSNAME-GLOB argument should be put in quotes.

    If CLASSNAME-GLOB is not specified, then all classes in the specified
    namespaces are counted (same as when specifying CLASSNAME-GLOB as "*").

    For example, "pywbem_*" returns classes whose name begins with "PyWBEM_",
    "pywbem_", etc. "*system*" returns classes whose names include the case
    insensitive string "system".

    This command can take a long time to execute since it potentially
    enumerates all instance names for all classes in all namespaces.
    """
    context.execute_cmd(lambda: cmd_instance_count(context, classname, options))


####################################################################
#
#  cmd_instance_<action> processors
#
####################################################################

def get_instancename(context, instancename, options):
    """
    Common function to get the instancename from either the input or the user.
    If the instance name replaces the keys with ".?" execute the console
    prompt to select the instance name.

    Returns:
     CIMInstanceName with namespace retrieved either from the namespace option
     in options dictionary or the connection default_namespace.
    """
    try:
        # if the keys component is the character ? execute select
        if instancename.endswith(".?"):
            instancename = instancename[:-2]
            ns = options.get('namespace', context.conn.default_namespace)

            try:
                instancepath = pick_instance(context, instancename,
                                             namespace=ns)
                return instancepath

            except ValueError:
                click.echo('Function aborted')
                return None

        else:
            instancepath = parse_wbemuri_str(instancename,
                                             options['namespace'])
        return instancepath

    except ValueError as ve:
        raise click.ClickException("%s: %s" % (ve.__class__.__name__, ve))


def cmd_instance_get(context, instancename, options):
    """
    Get and display an instance of a CIM Class.

    Gets the instance defined by instancename argument and displays in output
    format defined.

    If the wildcard key is used (CLASSNAME.?), pywbemcli presents a list of
    instances to the console from which one can be picked to get from the
    server and display.
    """
    instancepath = get_instancename(context, instancename, options)
    if instancepath is None:
        return

    try:
        instance = context.conn.GetInstance(
            instancepath,
            LocalOnly=options['local_only'],
            IncludeQualifiers=options['include_qualifiers'],
            IncludeClassOrigin=options['include_classorigin'],
            PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, instance, context.output_format)

    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))


def cmd_instance_delete(context, instancename, options):
    """
        If option interactive is set, get instances of the class defined
        by instance name and allow the user to select the instance to
        delete.
        Otherwise attempt to delete the instance defined by instancename
    """
    instancepath = get_instancename(context, instancename, options)
    if instancepath is None:
        return

    try:
        context.conn.DeleteInstance(instancepath)

        if context.verbose:
            click.echo('Deleted %s' % instancepath)

    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))


def cmd_instance_create(context, classname, options):
    """Create an instance and submit to wbemserver.
       If successful, this operation returns the new instance name. Otherwise
       it raises an exception
    """
    try:
        class_ = context.conn.GetClass(
            classname, namespace=options['namespace'], LocalOnly=False)
    except CIMError as ce:
        if ce.status_code == CIM_ERR_NOT_FOUND:
            ns = options['namespace'] or context.conn.default_namespace
            raise click.ClickException('CIMClass: "%s" does not exist in '
                                       'namespace "%s" in WEB '
                                       'server: %s.'
                                       % (classname, ns, context.conn))

        raise click.ClickException('Exception %s' % ce)
    except Error as er:
        raise click.ClickException('Exception %s' % er)

    properties = options['property']

    # properties is a tuple of name,value pairs
    new_inst = create_ciminstance(class_, properties)

    if options['verify']:
        context.spinner.stop()
        click.echo(new_inst.tomof())
        if not verify_operation("Execute CreateInstance", msg=True):
            return
    try:
        name = context.conn.CreateInstance(new_inst,
                                           namespace=options['namespace'])

        context.spinner.stop()
        click.echo("%s" % name)
    except Error as er:
        raise click.ClickException("Server Error creating instance. Exception: "
                                   "%s: %s" % (er.__class__.__name__, er))


def cmd_instance_modify(context, instancename, options):
    """
    Build an instance defined by the options and submit to wbemserver
    as a ModifyInstance method.

    In order to make a correct instance, this method first gets the
    corresponding class and uses that as the template for creating the intance.
    The class provides property names, types, array flags, etc. to assure
    that the instance is correctly built.

    If successful, this operation returns nothing.
    """
    # This function resolves any issues between namespace in instancename and
    # the namespace option.
    instancepath = get_instancename(context, instancename, options)
    if instancepath is None:
        return

    try:
        class_ = context.conn.GetClass(
            instancepath.classname, LocalOnly=False)
    except CIMError as ce:
        if ce.status_code == CIM_ERR_NOT_FOUND:
            raise click.ClickException('CIMClass: %r does not exist in WEB '
                                       'server: %s'
                                       % (instancepath.classname,
                                          context.conn.url))

        raise click.ClickException('Exception %s' % ce)
    except Error as er:
        raise click.ClickException('Exception %s' % er)

    property_list = resolve_propertylist(options['propertylist'])

    # properties is a tuple of name,value pairs
    modified_inst = create_ciminstance(class_, options['property'])

    modified_inst.path = instancepath

    if options['verify']:
        context.spinner.stop()
        click.echo(modified_inst.tomof())
        if not verify_operation("Execute ModifyInstance", msg=True):
            return

    try:
        context.conn.ModifyInstance(modified_inst,
                                    PropertyList=property_list)
    except Error as er:
        raise click.ClickException("Server Error modifying instance. "
                                   "Exception: %s: %s" %
                                   (er.__class__.__name__, er))


def cmd_instance_invokemethod(context, instancename, methodname,
                              options):
    """Create an instance and submit to wbemserver"""
    instancepath = get_instancename(context, instancename, options)
    if instancepath is None:
        return

    try:
        process_invokemethod(context, instancepath, methodname, options)
    except Exception as ex:
        raise click.ClickException("%s: %s" % (ex.__class__.__name__, ex))


def get_filterquerylanguage(options):
    """
    Get the filterquery language based on what is in the filterquery option
    and the filterquerylanguage options.
    If filterquery exists but filterquerylanguage does not, use DMTF as
    the filter query language.
    if filterquery does not exist but filterquerylanguage does, just return it
    """
    if options['filter_query']:
        fql = options['filter_query_language'] or 'DMTF:FQL'
    else:
        fql = options['filter_query_language']
    return fql


def cmd_instance_enumerate(context, classname, options):
    """
    Enumerate CIM instances or CIM instance names

    """
    try:
        if options['names_only']:
            results = context.conn.PyWbemcliEnumerateInstancePaths(
                ClassName=classname,
                namespace=options['namespace'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt)
        else:
            results = context.conn.PyWbemcliEnumerateInstances(
                ClassName=classname,
                namespace=options['namespace'],
                LocalOnly=options['local_only'],
                IncludeQualifiers=options['include_qualifiers'],
                DeepInheritance=options['deep_inheritance'],
                IncludeClassOrigin=options['include_classorigin'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt,
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, context.output_format,
                            summary=options['summary'], sort=True)

    except (Error) as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))
    except ValueError as ve:
        raise click.ClickException('instance enumerate failed because '
                                   'FilterQuery not allowed with traditional '
                                   'EnumerateInstance. --use-pull: '
                                   '%s. Exception: %s: %s' %
                                   (context.use_pull,
                                    ve.__class__.__name__, ve))


def cmd_instance_references(context, instancename, options):
    """Execute the references request operation to get references for
       the classname defined. This may be either interactive or if the
       interactive option is set or use the instancename directly.

       If the interactive option is selected, the instancename MUST BE
       a classname.
    """
    instancepath = get_instancename(context, instancename, options)
    if instancepath is None:
        return

    try:
        if options['names_only']:
            results = context.conn.PyWbemcliReferenceInstancePaths(
                instancepath,
                ResultClass=options['result_class'],
                Role=options['role'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt)
        else:
            results = context.conn.PyWbemcliReferenceInstances(
                instancepath,
                ResultClass=options['result_class'],
                Role=options['role'],
                IncludeQualifiers=options['include_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt,
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, context.output_format,
                            summary=options['summary'], sort=True)

    except (Error) as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))
    except ValueError as ve:
        raise click.ClickException('instance references failed because '
                                   'FilterQuery not allowed with traditional '
                                   'References. --use-pull: '
                                   '%s. Exception: %s: %s' %
                                   (context.use_pull,
                                    ve.__class__.__name__, ve))


def cmd_instance_associators(context, instancename, options):
    """
    Execute the references request operation to get references for
    the classname defined
    """
    instancepath = get_instancename(context, instancename, options)
    if instancepath is None:
        return

    try:
        if options['names_only']:
            results = context.conn.PyWbemcliAssociatorInstancePaths(
                instancepath,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt)
        else:
            results = context.conn.PyWbemcliAssociatorInstances(
                instancepath,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'],
                IncludeQualifiers=options['include_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt,
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, context.output_format,
                            summary=options['summary'], sort=True)

    except (Error) as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))
    except ValueError as ve:
        raise click.ClickException('instance associators failed because '
                                   'FilterQuery not allowed with traditional '
                                   'Associators. --use-pull: '
                                   '%s. Exception: %s: %s' %
                                   (context.use_pull,
                                    ve.__class__.__name__, ve))


def cmd_instance_count(context, classname, options):
    """
    Get the number of instances of each class in the namespace
    """
    if classname is None:
        classname = '*'
    # Create list of namespaces from the option or from all namespaces
    if options['namespace']:
        ns_names = options['namespace']
    else:
        try:
            ns_names = context.wbem_server.namespaces
            ns_names.sort()
        except CIMError as ce:
            # allow processing to continue if no interop namespace
            if ce.status_code == CIM_ERR_NOT_FOUND:
                click.echo('WARNING: %s' % ce)
                ns_names = [context.conn.default_namespace]

    ns_cln_tuples = []  # a list of tuples of namespace, classname
    for namespace in ns_names:
        # Get all classes in Namespace
        try:
            classnames = context.conn.EnumerateClassNames(
                DeepInheritance=True,
                namespace=namespace)
        except Error as er:
            raise click.ClickException("%s: %s" % (er.__class__.__name__, er))

        if classnames:
            classlist = filter_namelist(classname, classnames, ignore_case=True)
            cl_tup = [(namespace, cln) for cln in classlist]
            ns_cln_tuples.extend(cl_tup)

    # sort since normal output for this command is  namespace, classname
    # alphabetic order.
    ns_cln_tuples.sort(key=lambda tup: (tup[0], tup[1]))

    display_data = []
    for tup in ns_cln_tuples:
        ns = tup[0]
        cln = tup[1]
        # Try block allows issues where enumerate does not properly execute
        # The totals may be wrong but at least it gets what it can.
        # This accounts for issues with some servers where there
        # are providers that return errors from the enumerate.
        try:
            inst_names = context.conn.EnumerateInstanceNames(cln, namespace=ns)
        except Error as er:
            click.echo('Server Error %s with %s:%s. Continuing' %
                       (er, ns, cln), err=True)

        # Sum the number of instances with the defined classname.
        # this counts only classes with that specific classname and not
        # subclasses
        clnl = cln.lower()
        count = sum(1 for inst_name in inst_names
                    if (inst_name.classname.lower() == clnl))

        if count != 0:
            display_tuple = (ns, cln, count)
            display_data.append(display_tuple)

    # If sort set, resort by count size
    if options['sort']:
        display_data.sort(key=lambda x: x[2])

    headers = ['Namespace', 'Class', 'count']
    rows = []
    if display_data:
        for item in display_data:
            rows.append([item[0], item[1], item[2]])

    context.spinner.stop()
    click.echo(format_table(rows, headers,
                            title='Count of instances per class',
                            table_format=context.output_format))


def cmd_instance_query(context, query, options):
    """Execute the query defined by the inputs"""

    try:
        results = context.conn.PyWbemcliQueryInstances(
            options['query_language'],
            query,
            namespace=options['namespace'],
            MaxObjectCount=context.pull_max_cnt)

        display_cim_objects(context, results, context.output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))
